import os
from typing import Optional
from pathlib import Path

import requests
from tqdm import tqdm


def _get_config_path(program_name="spine") -> Path:
    """Get OS-agnostic path for program configuration."""

    home = Path.home().absolute()

    if os.name == "posix":
        # Linux, macOS, and other UNIX-like systems
        config_path = home / ".config" / program_name
    elif os.name == "nt":
        # Windows
        local_app_data = Path(os.getenv("LOCALAPPDATA"))  # For machine-local configurations
        config_path = local_app_data / program_name
    else:
        # Fallback for unknown OS
        config_path = home / f".{program_name}"

    config_path.mkdir(parents=True, exist_ok=True)
    return config_path


def download_file_with_progress(url: str, local_filename: Optional[Path] = None) -> Path:
    """Download file if it does not exist, and save persistently to local file. Return its path either way."""
    if local_filename is None:
        local_filename = _get_config_path() / '/'.join(url.split("/")[-2:])

    if local_filename.exists():
        return local_filename

    local_filename.parent.mkdir(exist_ok=True, parents=True)

    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get("content-length", 0))

    block_size = 1024

    print(f"Downloading {local_filename}...")
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
    with open(local_filename, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()

    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        print("ERROR, something went wrong")
    else:
        print(f"Downloaded {local_filename}")

    return local_filename


def main():
    # Example usage
    url = "https://github.com/LiquidFun/Spine/releases/download/onnx_models_1.0.0/2023-11-06_Seg2_Unet_resnet152_896px.onnx"
    download_file_with_progress(url)


if __name__ == "__main__":
    main()



from typing import List, Union

import numpy as np


class LineSegment:
    def __init__(self, points: Union[np.ndarray, List[np.ndarray]], *, name=None):
        self.points = np.array(points)
        self.name = name



import math
import re
from collections import defaultdict
from pathlib import Path

import pandas as pd
import yaml


def last_key(config, location):
    if location == "":
        return config.split(".")[-1]
    rest = location.split(".", 1)[1] if "." in location else ""
    return last_key(config[location.split(".")[0]], rest)


def maybe_int(string):
    try:
        return int(string)
    except ValueError:
        return string


def split_into_numbers_and_text_tuple(string):
    res = re.split(r"(\d+|\D+)", string)
    return list(map(maybe_int, res))


def textbf_if_max(value, max_value, overall_max_value):
    string = f"{value:.3f}"
    with_star = "$^{\star}$" if round(value, 3) == overall_max_value else ""
    return (r"\textbf{" + string + "}" if round(value, 3) == max_value else string) + with_star


def with_citation(string):
    citation = ""
    for key, val in term_to_citation.items():
        if key.lower() in string.lower():
            citation = r" \cite{" + val + "}"
    marker = ""
    if "timm-" in string:
        marker = r"$^{\dagger}$"

    return string.replace("_", r"\_") + marker + citation


term_to_citation = {
    "Unet": "Ronneberger_2015",
    "DeepLabV3Plus": "Chen2018",
    "DeepLabV3": "Chen2017",
    "MAnet": "Fan2020",
    "ResNet": "He2015",
    "FPN": "Lin2016",
    "PAN": "Li2018",
    "DenseNet": "Huang2016",
    "EfficientNet": "Tan2019",
    "ResNeSt": "Zhang2020",
    "ResNeXt": "Xie2016",
    "VGG": "Simonyan2014",
    "Linknet": "Chaurasia2017",
    "mit": "Xie2021",
    "Res2Net": "Gao2021",
    "RegNetX": "Radosavovic2020",
    "GerNet": "VanHilten2021",
    "MobileOne": "Vasu2022",
    # "CrossEntropyLoss": "Goodfellow2016",
}


def main():
    csv = pd.read_csv(Path(__file__).absolute().parent / "task1_results/runs_all.csv")
    table = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: (0, 0))))
    overall_max_dice = 0
    overall_max_iou = 0
    for row in csv.itertuples():
        with open(row.ConfigPath) as config_file:
            config = yaml.load(config_file, yaml.Loader)
            model = last_key(config, "model.init_args.model.class_path")
            encoder = last_key(config, "model.init_args.model.init_args.encoder_name")
            loss = last_key(config, "model.init_args.loss.class_path")
            augmentation = False
            try:
                last_key(config, "model.init_args.augmentation")
                augmentation = True
            except (AttributeError, KeyError):
                pass
            # print(augmentation)
            dice = row.val_f1
            iou = row.val_iou
            if not math.isnan(dice) and not math.isnan(iou):
                prev_dice, prev_iou = table[model][encoder][loss]
                table[model][encoder][loss] = (max(prev_dice, dice), max(prev_iou, iou))
            overall_max_dice = max(overall_max_dice, round(dice, 3))
            overall_max_iou = max(overall_max_iou, round(iou, 3))
            # print(model, encoder, loss)

    # for model in table:
    #     print(model)
    #     for encoder in table[model]:
    #         print(f"\t{encoder}")
    #         for loss in table[model][encoder]:
    #             print(f"\t\t{loss}: {table[model][encoder][loss]}")
    #         print()

    print(r"\hline")
    print(" & ".join(["Section", "Architecture", "Encoder", "Loss", "\Ac{DSC}", "\Ac{IoU}"]) + r" \\\hline\hline")
    for model in table:
        max_dice = 0
        max_iou = 0
        for encoder in sorted(table[model], key=split_into_numbers_and_text_tuple):
            for loss in table[model][encoder]:
                max_dice = max(max_dice, round(table[model][encoder][loss][0], 3))
                max_iou = max(max_iou, round(table[model][encoder][loss][1], 3))

        for encoder in sorted(table[model], key=split_into_numbers_and_text_tuple):
            # if "JaccardLoss" in table[model][encoder] and "DiceLoss" in table[model][encoder]:
            for loss in table[model][encoder]:
                dice, iou = table[model][encoder][loss]
                line = [
                    r"\ref{sec:slice-based-segmentation}",
                    with_citation(model),
                    with_citation(encoder),
                    with_citation(loss),
                    textbf_if_max(dice, max_dice, overall_max_dice),
                    textbf_if_max(iou, max_iou, overall_max_iou),
                ]
                joined = " & ".join(line) + r" \\"
                print(joined)
        print(r"\hline")


if __name__ == "__main__":
    main()



from collections import Counter
from pathlib import Path

import matplotlib
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from spine_segmentation.resources.paths import LOGS_PATH
from spine_segmentation.spine_types.labels import get_labels_for_n_classes
from spine_segmentation.visualisation.color_palettes import get_wk_bs_palette


def plot_accuracy(correct_counter: Counter[int, int], appeared_counter: Counter[int, int], V: int, N: int):
    # Calculate accuracy for each class

    matplotlib.rc("font", size=21)

    labels = get_labels_for_n_classes(49)
    # to_x = lambda x: x + (x > 12) * 1 + (x > 36) * 1
    to_x = lambda x: x  # + (x > 12) * 1 + (x > 36) * 1

    for i in set(correct_counter):
        if i not in range(1, 50):
            del correct_counter[i]
    for i in set(appeared_counter):
        if i not in range(1, 50):
            del appeared_counter[i]

    accuracy = {to_x(k): correct_counter[k] / appeared_counter[k] for k in sorted(appeared_counter.keys())}

    # Prepare data for seaborn
    data = pd.DataFrame(list(accuracy.items()), columns=["Class", "Accuracy"])  # , index=map(to_x, range(49)))

    palette = get_wk_bs_palette()
    del palette[0]
    colors = np.array([list(c.floats()) for c in palette.values()])
    colors = np.tile(colors.T, 100).T[:49]
    is_thoracic = np.array([l.startswith("T") for l in labels])
    # colors[is_thoracic] *= 0.7
    # data["Hue"] = [v.ints() for k, v in palette.items()]
    # palette = {int(k): v.ints() for k, v in palette.items()}
    # Create seaborn barplot
    plt.figure(figsize=(10, 6))
    # sns.barplot(x="Class", y="Accuracy", dodge=False, data=data, palette=colors)
    bars = plt.bar(data["Class"], data["Accuracy"], color=colors)
    plt.margins(x=0.003)
    plt.legend(bars[:2], ["Vertebra", "Intervertebral disc"], loc="lower left")
    # plt.bar(range(50), accuracy.values())
    # plt.title(f"Multiclass Segmentation Accuracy per Vertebra and Intervertebral Disc for FOV: V={V} and N={N}")
    # print(len(labels), len(range()))
    # fontsizes = [20 if i % 2 == 1 else 10 for i in range(49)]
    vertebra_labels = [label if i % 2 == 0 else "" for i, label in enumerate(labels)]
    plt.xticks(list(map(to_x, range(1, 50))), vertebra_labels, rotation=90)
    plt.ylabel("Accuracy")

    yticks = np.linspace(0, 1, 21)
    percent = [f"{int(y * 100)}%" for y in yticks]
    plt.yticks(yticks, percent)
    ax = plt.gca()  # Get current axes

    for i, line in enumerate(ax.get_ygridlines()):
        if i % 2 == 0:  # for even indices
            line.set(alpha=0.8)
            line.set_linestyle("-")
            line.set_linewidth(1)
        else:  # for odd indices
            line.set_linestyle("-")
            line.set_linewidth(1)
            line.set(alpha=0.3)

    # plt.grid(axis="y", color=[0, 0, 0], alpha=0.2)
    plt.grid(axis="y", color=[0, 0, 0])
    plt.tight_layout(pad=0)
    plot_path = Path(__file__).parent / "plots" / "accuracy_per_class" / f"PerROIAccuracyV{V}.png"
    plot_path.parent.mkdir(exist_ok=True)
    plt.savefig(plot_path, dpi=300)
    plt.savefig(f"/tmp/perroiaccuracy/{plot_path.name}", dpi=300)
    plt.show()


def main():
    # log_dirs = ["2023-11-06_232432", "2023-11-06_232433", "2023-11-06_232435"]
    # log_dirs = ["2023-11-06_232433"]

    # 30k but bad appeared counter
    # log_dirs = [
    #     ("2023-11-07_210729", 15),
    #     ("2023-11-07_210805", 10),
    #     ("2023-11-07_210824", 5),
    # ]

    # Good nice data
    log_dirs = [
        ("2023-11-08_134517", 20),
        ("2023-11-08_134323", 15),
        ("2023-11-08_134534", 10),
        ("2023-11-08_134551", 5),
    ]

    # Something fixed here ?
    # log_dirs = [
    #     ("2023-11-08_185546", 15),
    #     ("2023-11-08_185614", 10),
    #     ("2023-11-08_185820", 5),
    # ]

    # Test set with new segmentation and new inst segmentation
    log_dirs = [
        ("2023-11-09_044759", 15),
        ("2023-11-09_044842", 10),
        ("2023-11-09_044825", 5),
    ]

    log_dirs = [("2023-11-09_012718", 8)]

    log_dirs = [
        ("2023-11-09_144956", 5),
        ("2023-11-09_144959", 7),
        ("2023-11-09_145015", 10),
        ("2023-11-09_145032", 15),
    ]

    for log_dir_name, V in log_dirs:
        log_path = LOGS_PATH / log_dir_name
        npz_path = log_path / "stats.npz"
        npz = np.load(npz_path, allow_pickle=True)

        stats = npz["stats"]

        correct_counter = Counter()
        correct_no_edges_counter = Counter()
        appeared_counter = Counter()
        appeared_no_edges_counter = Counter()
        dice_scores = []

        corr = []
        corr_no_edges = []

        print(len(stats))
        for pat in stats:
            if "correct_counter" in pat and "correct_no_edges_counter" in pat:
                correct_counter += pat["correct_counter"]
                correct_no_edges_counter += pat["correct_no_edges_counter"]
                appeared_counter += pat["appeared_counter"]
                appeared_no_edges_counter += pat["appeared_no_edges_counter"]
            if "f1" in pat:
                dice_scores.append(pat["f1"])
                corr_no_edges.append(pat["correct_no_edges"])
                corr.append(pat["correct"])
        del appeared_counter[0]
        plot_accuracy(correct_no_edges_counter, appeared_no_edges_counter, V=V, N=sum(appeared_counter.values()))

        dice_scores = np.array(dice_scores)

        # print(correct_counter)
        print(sorted(correct_no_edges_counter.items()))
        print(sorted(appeared_counter.items()))
        print(f"Dice: (V={V})", np.mean(dice_scores))
        print(f"Correct: (V={V})", np.mean(corr))
        print(f"Correct no edges: (V={V})", np.mean(corr_no_edges))
        print()


if __name__ == "__main__":
    main()



from collections import Counter, defaultdict

from matplotlib import pyplot as plt

from spine_segmentation.resources.preloaded import get_measure_statistics


def plot_nan_counts_by_column(stats):
    nan_counts = stats.dicom_metadata.isna().sum()
    # nan_counts = nan_counts[nan_counts > 0]
    nan_counts = nan_counts.sort_values(ascending=False)
    # print(nan_counts.to_string())
    nan_counts.plot.bar()
    plt.show()


def count_column_prefixes(stats):
    counts = Counter(column.split(";")[0] for column in stats.nan_filtered_statistics.columns)
    print(*counts.items(), sep="\n")


def print_metadata(stats):
    metadata = stats.dicom_metadata
    metadata = stats.statistics
    subset = metadata[:10]
    subset.loc[-1] = metadata.isna().sum()
    subset.loc[-2] = metadata.nunique()
    subset = subset.sort_index()
    # table = pd.concat([subset, nan_counts])
    print(subset.T.to_string())
    print(subset.T.to_csv("metadata.csv"))

    print(sorted(stats.statistics.columns.str.extract(r";(.*)")[0].unique()))

    print(subset.T[subset.T[-2] != 1].to_string())


def print_stat_columns(stats):
    keys = sorted(stats.statistics.columns.str.extract(r";(.*)")[0].unique())
    multikeys = defaultdict(set)
    prefixes = defaultdict(set)
    for key in keys:
        prefix, *middle, postfix = key.split("_")
        multikeys["_".join(middle)].add(postfix)
        prefixes["_".join(middle)].add(prefix)

    for key, value in multikeys.items():
        print(f"{{{','.join(prefixes[key])}}}_{key}_{{{','.join(value)}}}")

    # print(*, sep="\n")


def main():
    stats = get_measure_statistics()
    plot_nan_counts_by_column(stats)
    print_metadata(stats)
    print_stat_columns(stats)
    # count_column_prefixes(stats)


if __name__ == "__main__":
    main()



from collections import defaultdict

import numpy as np

# 192 px
# STATS_PATH = "logs/2023-09-25_194839/stats.npz"

# 416 px
STATS_PATH = "logs/2023-09-26_094648/stats.npz"

# 320 px
# STATS_PATH = "logs/2023-09-26_105632/stats.npz" # No fix top
# STATS_PATH = "logs/2023-09-26_120239/stats.npz"


def main():
    stats = np.load(STATS_PATH, allow_pickle=True)["stats"]
    accumulated = defaultdict(int)
    for i, stat in enumerate(stats):
        if stat == "Error":
            print("Error")
            # for key in ["f1", "correct", "correct_no_edges"]:
            #     accumulated[key] += 0
        else:
            print(stat["correct_no_edges"])
            for key in stat:
                if key in ["f1", "correct", "correct_no_edges"]:
                    accumulated[key] += stat[key]
        # print(i, stat.keys())
    # correct = sum(stat["correct"] for stat in stats)
    # correct_no_edges = sum(stat["correct_no_edges"] for stat in stats)
    for key in accumulated:
        accumulated[key] /= len(stats)
        print(key, accumulated[key])


if __name__ == "__main__":
    main()



"""Module which initializes blender scene with objects and parses arguments

This script is meant to be used in conjunction with blender.
It can be added with the -P flag when running blender on
the command line, blender will run this script once the scene has been initialized.
Note that this is might be a different python version, using the global
pip packages instead of the local venv, so ideally no external
packages should be used (such as numpy).
"""

import argparse
import math
import sys
from functools import partial
from pathlib import Path

cameras = []


def load_obj(path):
    """Loads .obj file and returns the object"""
    name = Path(path).name.replace(".obj", "")
    if not Path(path).exists():
        print(f"File {path} does not exist! Skipping it!")
        return None
    bpy.ops.import_scene.obj(filepath=path)
    return bpy.context.selected_objects[-1]


def select(obj, active=False):
    obj.select_set(True)
    if active:
        bpy.context.view_layer.objects.active = obj


def hide(obj, viewport=False, selection=False, render=False):
    obj.hide_viewport = viewport
    obj.hide_render = render
    obj.hide_select = selection


def make_obj_smooth(obj, iterations=5, factor=2):
    """Adds smoothing modifier in Blender"""
    smoothing = obj.modifiers.new(name="Smooth", type="SMOOTH")
    smoothing.iterations = iterations
    smoothing.factor = factor
    bpy.ops.object.shade_smooth()
    recalculate_normals(obj)


def recalculate_normals(obj):
    bpy.ops.object.select_all(action="DESELECT")
    select(obj, True)
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.select_all(action="SELECT")
    bpy.ops.mesh.normals_make_consistent(inside=False)
    bpy.ops.object.editmode_toggle()


def prepare_rendering():
    bpy.context.scene.render.engine = "CYCLES"
    bpy.context.scene.cycles.samples = 256
    # Reenable this when using a different blender version
    bpy.context.scene.cycles.use_denoising = False
    # bpy.context.scene.cycles.denoiser = "OPTIX"
    bpy.context.scene.render.resolution_x = 800
    bpy.context.scene.render.resolution_y = 2000
    bpy.context.scene.render.resolution_percentage = 100


def set_background_to_transparent():
    bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (0, 0, 0, 1)
    bpy.context.scene.render.film_transparent = True


def set_up_default_preferences():
    # set viewport background to white
    bpy.context.preferences.themes["Default"].view_3d.space.gradients.high_gradient = (1, 1, 1)
    # turn of splash screen
    bpy.context.preferences.view.show_splash = False
    # set line thickness
    # bpy.context.preferences.ui_line_width = "thick"


def add_light_plane():
    bpy.context.view_layer.active_layer_collection = bpy.context.view_layer.layer_collection.children["Collection"]
    bpy.ops.mesh.primitive_plane_add()
    plane = bpy.data.objects["Plane"]
    plane.name = "LightPlane"
    # hide(plane, viewport=True)
    plane.scale = (22, 80, 1)
    plane.location = (0, 20, 40)

    mat_name = "LightMat"
    mat = bpy.data.materials.new(mat_name)
    mat.use_nodes = True
    mat.node_tree.nodes.new(type="ShaderNodeEmission")
    mat.node_tree.nodes["Emission"].inputs[1].default_value = 5
    inp = mat.node_tree.nodes["Material Output"].inputs["Surface"]
    outp = mat.node_tree.nodes["Emission"].outputs["Emission"]
    mat.node_tree.links.new(inp, outp)
    plane.active_material = mat


def remove_default_objects():
    for object_name in ["Cube", "Lamp", "Light"]:
        if object_name in bpy.data.objects:
            current_object = bpy.data.objects[object_name]
            bpy.data.objects.remove(current_object)


def deg2rad(angle_degrees: float) -> float:
    return angle_degrees * math.pi / 180


def rad2deg(angle_radians: float) -> float:
    return angle_radians / math.pi * 180


def set_up_camera():
    cor_camera = bpy.data.objects["Camera"]
    cor_camera.location = (24, 0, 0)
    cor_camera.rotation_euler = (deg2rad(90), 0, deg2rad(90))
    cor_camera.name = "CoronalCamera"
    cameras.append(cor_camera)
    # hide(camera, viewport=True)

    bpy.ops.object.camera_add(location=(17, 17, 0))
    corsag_camera = bpy.context.active_object
    corsag_camera.rotation_euler = (deg2rad(90), 0, deg2rad(135))
    corsag_camera.name = "CoroSagittalCamera"
    cameras.append(corsag_camera)

    bpy.ops.object.camera_add(location=(0, 24, 0))
    sag_camera = bpy.context.active_object
    sag_camera.rotation_euler = (deg2rad(90), 0, deg2rad(180))
    sag_camera.name = "SagittalCamera"
    cameras.append(sag_camera)

    bpy.context.scene.camera = corsag_camera


class AddPlaneAction(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        def parse_plane(plane_as_str: str):
            """
            >>> parse_plane("Th12=[(1, 2, 3), (2, 3, 4), (2.2, 4.2, 5.2), (1e2, -2e1, -2.3e-4)]")
            ('Th12', [(1.0, 2.0, 3.0), (2.0, 3.0, 4.0), (2.2, 4.2, 5.2), (100.0, -20.0, -0.00023)])
            """
            import re

            plane_as_str_no_whitespace = re.sub(r"\s+", "", plane_as_str)
            match = re.fullmatch(r"(\w+)=\[(.*)]", plane_as_str_no_whitespace)
            if match:
                name = match.group(1)
                # number = r"[+-]?((\d+\.?\d*)?|\.\d+)([eE]-?\d+)?"
                number = r"[-+0-9e.]+"
                tuple_regex = rf"\(({number},{number},{number})\)"
                coordinates = re.findall(tuple_regex, match.group(2))
                coordinates = [tuple(map(float, coord.split(","))) for coord in coordinates]
                return name, coordinates
            return None

        if not hasattr(args, "planes"):
            setattr(args, "planes", [])
        for value in values:
            plane_tuple = parse_plane(value)
            if plane_tuple:
                args.planes.append(plane_tuple)


class AddObjectAction(argparse.Action):
    def __init__(self, smooth=False, *args, **kwargs):
        self.smooth = smooth
        super().__init__(*args, **kwargs)

    def __call__(self, parser, args, values, option_string=None, smooth=False):
        if not hasattr(args, "objects"):
            setattr(args, "objects", [])
        for path in values:
            obj = load_obj(path)
            if obj is not None:
                # obj.rotation_euler = (0, 0, 0)
                obj.show_name = True
                if self.smooth:
                    make_obj_smooth(obj)
                bpy.ops.object.origin_set(type="ORIGIN_GEOMETRY", center="MEDIAN")
                args.objects.append(obj)


def get_areas_by_type(context, type_):
    return [a for a in context.screen.areas if a.type == type_]


def show_names_in_current_screen():
    for view3d_area in get_areas_by_type(bpy.context, "VIEW_3D"):
        view_space = view3d_area.spaces[0]
        view_space.show_only_render = False
        view_space.show_floor = False
        view_space.show_axis_x = False
        view_space.show_text = False
        view_space.show_axis_y = False
        view_space.show_axis_z = False
        view_space.cursor_location = (0, 0, 1000)


def make_intersection(for_obj, intersecting_obj):
    modifier_name = f"{for_obj.name}_{intersecting_obj.name}_intersect_modifier"
    mask_modifier = for_obj.modifiers.new(type="BOOLEAN", name=modifier_name)
    mask_modifier.object = intersecting_obj
    mask_modifier.operation = "INTERSECT"
    bpy.context.view_layer.objects.active = for_obj
    # Not needed because on export apply_modifiers=True is used, however, it is necessary
    # so that normals can be recalculated.
    bpy.ops.object.modifier_apply(apply_as="DATA", modifier=modifier_name)
    recalculate_normals(for_obj)


def apply_boolean_mask_to_spine_object():
    for obj_name in ["Spinal_canal", "Vertebrae", "Intervertebral_discs"]:
        obj = bpy.data.objects.get(obj_name, None)
        mask_cube = bpy.data.objects.get("AxisParallelCuboid1", None)
        if mask_cube is None or obj is None:
            continue
        make_intersection(obj, mask_cube)


class AddCollectionAction(argparse.Action):
    def __init__(self, hidden=False, *args, **kwargs):
        self.hidden = hidden
        super().__init__(*args, **kwargs)

    def __call__(self, parser, args, values, option_string=None):
        name = "collections" + ("_hidden" if self.hidden else "")
        if not hasattr(args, name):
            setattr(args, name, [])
        for value in values:
            collection = bpy.data.collections.new(value)
            bpy.context.scene.collection.children.link(collection)
            bpy.context.view_layer.active_layer_collection = bpy.context.view_layer.layer_collection.children[value]
            getattr(args, name).append(collection)


class ExportToAction(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        if not hasattr(args, "export_to"):
            setattr(args, "export_to", [])
        args.export_to.append(Path(values))


class RenderToAction(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        if not hasattr(args, "render_to"):
            setattr(args, "render_to", [])
        args.render_to.append(Path(values))


def disable_hidden_collections(parsed_args):
    if hasattr(parsed_args, "collections_hidden"):
        pass
        # view_layer = bpy.context.scene.view_layers["View Layer"]
        # hidden_collection_names = set(col.name for col in parsed_args.collections_hidden) | {"Collection"}
        # for col in view_layer.layer_collection.children:
        #     col.exclude = col.name in hidden_collection_names
        # Alternatively to hide monitor:
        #   bpy.data.collections[col.name].hide_viewport = True
        # Alternatively to hide in current viewport:
        #   bpy.context.view_layer.layer_collection.children[col.name].hide_viewport = True


def export_colladas(parsed_args):
    for path in parsed_args.export_to:
        print(path)
        bpy.ops.wm.collada_export(filepath=path, apply_modifiers=True)


def render_and_save(parsed_args):
    for camera in cameras:
        bpy.context.scene.camera = camera
        if not hasattr(parsed_args, "render_to"):
            return
        bpy.ops.render.render()
        render_image = bpy.data.images["Render Result"]
        for path in parsed_args.render_to:
            render_image.save_render(filepath=str(path.with_stem(path.stem + camera.name)))


def main():
    remove_default_objects()
    set_up_camera()
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--object", nargs="+", action=AddObjectAction)
    arg_parser.add_argument("--object-smooth", nargs="+", action=partial(AddObjectAction, smooth=True))
    arg_parser.add_argument("--collection", nargs="+", action=AddCollectionAction)
    arg_parser.add_argument("--collection-hidden", nargs="+", action=partial(AddCollectionAction, hidden=True))
    # arg_parser.add_argument("--plane", nargs="+", action=AddPlaneAction)
    arg_parser.add_argument("--export-to", default=[], action=ExportToAction)
    arg_parser.add_argument("--render-to", default=[], action=RenderToAction)
    parsed_args = arg_parser.parse_args(sys.argv[sys.argv.index("--") + 1 :])
    prepare_rendering()
    set_background_to_transparent()
    set_up_default_preferences()
    add_light_plane()
    apply_boolean_mask_to_spine_object()
    disable_hidden_collections(parsed_args)
    print(parsed_args)
    export_colladas(parsed_args)
    render_and_save(parsed_args)
    # show_names_in_current_screen()
    # bpy.ops.outliner.show_one_level(open=False)


if __name__ == "__main__":
    # Internal Blender import, not a package. Import here so that doctests can still be run
    # as modules. Doctests run as doctests are not considered to be main, and therefore this import is added here.
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    import bpy

    main()



import concurrent.futures.process
import datetime
import random
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from loguru import logger
from matplotlib import pyplot as plt
from tqdm import tqdm

from spine_segmentation.resources.paths import PLOTS_PATH, RESULTS_PATH
from spine_segmentation.resources.preloaded import get_measure_statistics
from spine_segmentation.spine_types.spine import Spine


def _file_with_fallback(file: Path) -> Path:
    if file.exists():
        return file
    logger.info("cannot find file, fallback: check in resource path")
    file_fallback = Path(__file__).parent.joinpath(file).resolve()
    if file_fallback.exists():
        logger.info(f"found file in resource path {file_fallback}.")
        return file_fallback
    logger.info("did not find file in resource path.")
    return file


@dataclass
class AnnotatorBase(ABC):
    debug_dir: Path = None

    @abstractmethod
    def _process(self, spine: Spine):
        raise NotImplementedError()

    def process(self, spine: Spine):
        """ """
        # self._algorithm_validation()
        res = self._process(spine)
        return res


class AnnotatorSimple(AnnotatorBase, ABC):
    """Annotates spine rois by labeling their indices"""

    stats = None
    statistics = None
    available_properties = ["DIRECTION"]  # , "CENTER", "VOLUME"]
    properties_to_be_used = available_properties

    # ================================================================================
    #  Initialization
    # ================================================================================

    def __init__(self, seed=None, load=True, save_to=None):
        logger.info("Initializing Annotator; this may take a while")
        if self.__class__.statistics is None:
            self.__class__.stats = self._load_measure_statistics()
            self.__class__.statistics = self.stats.statistics
        self.variable_lookup = dict(zip(self.available_properties, range(100)))
        self.property_indices = [self.variable_lookup[key] for key in self.properties_to_be_used]
        self.all_namings = list(self.stats.roi_naming_lookup.keys())
        self.all_namings_no_duplicates = list(self.stats.roi_naming_lookup.keys())
        self.roi_naming_lookup = self.stats.roi_naming_lookup
        # self.vector_table = self.stats.filter_by_column(
        # required_substrings=["DIRECTION_"], excluded_substrings=["CANAL_"], reshape=True
        # )[:, :, None, :]

        # self.vector_table = self.stats.roi_diff_vectors[:, :, None, :]
        self.vector_table = self.stats.canal_direction_vectors[:, :, None, :]

        # if load:
        #     self.vector_table = get_vector_table()
        # else:
        #     self.vector_table = self._normalize(self._initialize_vector_lookup_table())
        if save_to is not None:
            np.savez_compressed(save_to, vector_table=self.vector_table)
            print(f"Saving pre-computed vector table to {save_to}")
        self.max_roi_index = max(self.roi_naming_lookup.values())
        if seed is not None:
            np.random.seed(seed)
        self._validate_init()

    def _validate_init(self) -> None:
        """Validates properties which are assumed in the rest of the class"""
        assert all(("-" in naming) == i % 2 for naming, i in self.roi_naming_lookup.items()), "Not all discs are odd"

    def _get_sorted_namings(self, discard_below_percentage=0.01) -> Tuple[List[str], List[str], Dict[str, int]]:
        # All namings sorted
        all_namings = set()
        counts = {}
        for column in self.statistics.columns:
            if any(part in column for part in self.variable_lookup.keys()):
                roi_name = column.split(";")[0]
                all_namings.add(roi_name)
                counts[roi_name] = self.statistics[column].count()
        all_namings -= {"spinal_canal", "unexpected_0"}
        all_namings = [n for n in all_namings if counts[n] / len(self.statistics) > discard_below_percentage]
        all_namings += {"S1"}
        [(n, counts[n] / len(self.statistics)) for n in all_namings]
        all_namings_sorted = []
        for current_part_of_spine in "CTLS":
            sort_by = lambda name: int(name[1:].split("-")[0]) + 0.5 * ("-" in name)
            all_namings_sorted.extend([n for n in sorted(all_namings, key=sort_by) if current_part_of_spine == n[0]])

        # Roi naming lookup
        roi_naming_lookup: Dict[str, int] = {}
        vertebra_index = 0
        for naming in all_namings_sorted:
            roi_naming_lookup[naming] = vertebra_index - ("-" in naming)
            vertebra_index += 2 * ("-" not in naming)

        # All namings no duplicates
        naming_occ = Counter(roi_naming_lookup.values())
        to_be_removed = [n for n in all_namings if naming_occ[roi_naming_lookup.get(n, 0)] >= 2 and "-S1" in n]
        all_namings_no_duplicates = list(filter(lambda n: n not in to_be_removed, all_namings_sorted))
        return all_namings_sorted, all_namings_no_duplicates, roi_naming_lookup

    def _load_measure_statistics(self) -> pd.DataFrame:
        return get_measure_statistics()

    def _initialize_vector_lookup_table(self) -> np.array:
        relevant_columns = []
        for column in self.statistics.columns:
            # if 'DIRECTION' in column or "POSITION" in column:
            if any(part in column for part in self.variable_lookup.keys()):
                if all(part not in column for part in ["DIFF", "PROB"]):
                    relevant_columns.append(column)
        index = -1
        coordinate_index_lookup = {"X": 0, "Y": 1, "Z": 2, "ALL": (0, 1, 2)}
        shape_4d_array = (
            len(self.statistics),
            max(self.roi_naming_lookup.values()) + 1,
            len(self.variable_lookup),
            3,
        )
        patient_vectors = np.zeros(shape_4d_array)
        for row_index, row in list(self.statistics[relevant_columns].iterrows()):
            if row_index % 1000 == 0:
                logger.info(f"Annotator at row: {row_index}/{len(self.statistics)}")
            for column, value in row.items():
                naming, variable = column.split(";")
                if variable.endswith("VOLUME"):
                    variable += "_ALL"
                if not pd.isna(value) and naming in self.roi_naming_lookup:
                    multi_index = (
                        row_index,
                        self.roi_naming_lookup[naming],
                        self.variable_lookup[variable.split("_")[-2]],
                        coordinate_index_lookup[variable.split("_")[-1]],
                    )
                    assert np.all(patient_vectors[multi_index] == 0), "Array is not zero!"
                    patient_vectors[multi_index] = value

        # Filter some rows which have too many zero rows:
        max_allowed_all_0_entries = 4
        indices_to_be_filtered = (patient_vectors == 0).all(axis=(2, 3)).sum(axis=1) > max_allowed_all_0_entries
        patient_vectors = np.delete(patient_vectors, indices_to_be_filtered, axis=0)

        return patient_vectors

    def _normalize(self, vector_table):
        """Input either all patients shape (8512, 52, 2, 3) or a single patient shape (52, 2, 3)"""

        def divide_by_max(variable: str):
            max_val = vector_table[..., self.variable_lookup[variable], :].max(axis=(1, 2))
            vector_table[..., self.variable_lookup[variable], :] /= max_val.reshape(-1, 1, 1)

        def divide_by_global_max(variable: str):
            max_val = vector_table[..., self.variable_lookup[variable], :].max(axis=(0, 1, 2))
            vector_table[..., self.variable_lookup[variable], :] /= max_val

        def subtract_min(variable: str):
            min_val = vector_table[..., self.variable_lookup[variable], :].min(axis=(1, 2))
            vector_table[..., self.variable_lookup[variable], :] -= min_val.reshape(-1, 1, 1)

        is_single_vector = len(vector_table.shape) == 3
        if is_single_vector:
            vector_table = vector_table[np.newaxis, ...]

        if "VOLUME" in self.available_properties:
            divide_by_max("VOLUME")
        if "CENTER" in self.available_properties:
            subtract_min("CENTER")
            divide_by_max("CENTER")

        if is_single_vector:
            vector_table = vector_table[0]

        return vector_table

    # ================================================================================
    #  Samples & validation
    # ================================================================================

    def _evaluate_sample(self, size: int, patient_index=None, start_at=None):
        if patient_index is None:
            patient_index = random.randint(0, len(self.statistics) - 1)
        sample, start_at = self._get_random_sample(patient_index, size, start_at)
        table_without_sample = self._get_table_without_sample(patient_index)
        match = self._find_best_match(table_without_sample, sample, is_disc=start_at % 2 == 1, correct=start_at)
        # print(f"{match=} {start_at=} {size=}")
        return self.roi_naming_lookup[match[0]], start_at
        # if self.roi_naming_lookup[match[0]] == start_at:
        #     correct[size] += 1
        # correct_roi[start_at] = (correct_roi[start_at][0] + (match == start_at), correct_roi[start_at][1] + 1)

    def evaluate_each_roi_on_patient(self, size: int, patient_index: int = None):
        off_by = defaultdict(int)

        with concurrent.futures.ProcessPoolExecutor() as executor:
            to = self.max_roi_index - size
            args = [[size] * to, [patient_index] * to, range(0, to)]
            for guess, actual in tqdm(executor.map(self._evaluate_sample, *args)):
                off_by[actual] = guess
        return off_by

    def _algorithm_validation(self, iterations=100):
        """Tests random samples from NAKO on the entire NAKO dataset

        Each test is validated without the current sample. The start position is taken at random (can be disc or
        vertebra), each size from 1 to about 30 is tried with #iterations (e.g. 100).
        """
        correct = defaultdict(int)
        correct_plus_minus_1 = defaultdict(int)
        correct_roi = defaultdict(lambda: (0, 0))

        date = datetime.date.today().strftime("%Y-%m-%d")
        model_description = "canal_direction"
        postfix = f"{date}_{model_description}"

        with concurrent.futures.ProcessPoolExecutor(max_workers=160) as executor:
            for size in range(1, self.max_roi_index):
                for guess, actual in tqdm(executor.map(self._evaluate_sample, [size] * iterations)):
                    correct[size] += guess == actual
                    correct_plus_minus_1[size] += abs(guess - actual) <= 2
                    correct_roi[actual] = (correct_roi[actual][0] + (guess == actual), correct_roi[actual][1] + 1)
                print(f"correct: {dict(correct)}/{iterations}")
                print(f"corr+-2: {dict(correct_plus_minus_1)}/{iterations}")
                table = pd.DataFrame(
                    {
                        "slice size": range(1, size + 1),
                        "correct (%)": correct.values(),
                        "correct ±2 (%)": correct_plus_minus_1.values(),
                    }
                )
                print(table.to_markdown(index=False))

        table.to_csv(RESULTS_PATH / f"roi_accuracy_{postfix}.csv")
        self._plot_roi_accuracy(correct_roi, postfix)
        self._plot_slice_accuracy(correct, correct_plus_minus_1, iterations, postfix)

    def _plot_slice_accuracy(self, correct: Dict, correct_plus_minus_1: Dict, iterations: int, postfix: str):
        plt.bar(
            correct_plus_minus_1.keys(),
            [p / iterations for p in correct_plus_minus_1.values()],
            label="accuracy ±2",
            color="orange",
        )
        plt.bar(correct.keys(), [p / iterations for p in correct.values()], label="accuracy", color="green")
        plt.legend()
        plt.title(f"Accuracy of Vector Annotator by slice size ({postfix})")
        plt.ylabel("Accuracy %")
        plt.xlabel("Slice size")
        plt.xticks(list(correct.keys()))
        plt.yticks(np.arange(0, 1.1, 0.1))
        plt.savefig(PLOTS_PATH / f"slice_accuracy_{postfix}.png")
        logger.info("Annotator validation complete")

    def _plot_roi_accuracy(self, correct_roi: Dict, postfix):
        keys, values = list(zip(*sorted(correct_roi.items())))
        values = [correct / total for correct, total in values]
        plt.bar(keys, values)
        plt.xticks(keys)
        plt.yticks(np.arange(0, 1.1, 0.1))
        plt.ylabel("% of correct ROIs")
        plt.xlabel("ROI index")
        plt.title(f"Accuracy of Vector Annotator by ROI index ({postfix})")
        plt.savefig(PLOTS_PATH / f"roi_accuracy_{postfix}.png")
        plt.show()

    def _get_random_sample(self, patient_index: int, size=None, start_at=None):
        if size is None:
            size = random.randint(3, self.max_roi_index)
        if start_at is None:
            start_at = random.randint(0, self.max_roi_index - size)
        vector = self.vector_table[patient_index, start_at : start_at + size][..., self.property_indices, :]
        return vector, start_at

    def _get_subsets_at(self, table, start_at, size) -> List[Tuple[np.ndarray, List[str]]]:
        """Returns a list of tuples of subsets and namings

        * In total there are 1 or 2 subsets. 2 are returned only when it is ambiguous whether the last element is S1
        * The namings are the corresponding naming associated with this subset
        """
        subset_no_s1 = table[:, start_at : start_at + size, self.property_indices, :]
        naming_no_s1 = self.all_namings_no_duplicates[start_at : start_at + size]
        subsets = [(subset_no_s1, naming_no_s1)]
        s1_index = self.roi_naming_lookup.get("S1", None)
        if s1_index is not None:
            for name, index in self.roi_naming_lookup.items():
                # Check for discs which lead to S1
                is_second_to_last_index = index + 1 == start_at + size - 1
                if "-S1" in name and is_second_to_last_index and s1_index != index + 1:
                    indices_with_s1 = list(range(start_at, start_at + size - 1)) + [s1_index]
                    naming_with_s1 = naming_no_s1[:-1] + ["S1"]
                    if len(naming_with_s1) >= 2:
                        naming_with_s1[-2] = "-".join(naming_with_s1[-2].split("-")[:1] + ["S1"])
                    # Workaround with ..., because [:, indices_with_s1, self.property_indices, :] produces
                    # shape (x, 3, 2) due to advanced indexing (when 2 lists or tuples are used then an 'and'
                    # conjunction is used between them
                    subsets.append((table[:, indices_with_s1][..., self.property_indices, :], naming_with_s1))

        for i, (subset, naming) in enumerate(subsets):
            indices_to_be_removed = (subset == 0).all(axis=(2, 3)).any(axis=1)
            subsets[i] = (np.delete(subset, indices_to_be_removed, axis=0), naming)
        # assert all(subset.shape == subsets[0][0].shape for subset, _ in subsets), "Subsets are not of the same shape!"
        assert all(all(name in self.all_namings for name in naming) for _, naming in subsets), "Unknown name!"
        return subsets

    def _get_table_without_sample(self, filter_index: int):
        table_without_sample = self.vector_table[[i for i in range(len(self.vector_table)) if i != filter_index]]
        assert len(table_without_sample) + 1 == len(self.vector_table)
        return table_without_sample

    # ================================================================================
    #  Matching
    # ================================================================================

    @abstractmethod
    def _find_best_match(self, table: np.ndarray, sample: np.ndarray, is_disc: bool, correct: int) -> List[str]:
        ...

    def _is_naming_incomplete(self, spine) -> bool:
        return any(roi.name == "" for roi in (spine.vertebras + spine.discs))



import time
from pathlib import Path
from typing import Iterable

import lightning.pytorch as pl
import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from spine_segmentation.datasets.sample import GTFormat, MetadataSample, MetadataTypes, get_pid_to_index_lookup
from spine_segmentation.datasets.path_helper import get_potential_data_dirs
from spine_segmentation.resources.paths import NAKO_DATASET_PATH


class MetadataPredictionDataset(Dataset):
    def __init__(self, data_dirs: Iterable[Path], transform=None, gt_type: MetadataTypes = ["weight", "size", "age"]):
        super().__init__()
        self.target_shape = (18, 320, 896)
        self.data_dirs = list(data_dirs)
        self.transform = transform
        self.gt_type = gt_type
        self.add_adjacent_slices = True
        self.slice_wise = True
        self.slice_count = (self.target_shape[0] - self.add_adjacent_slices * 6) if self.slice_wise else 1

    def __len__(self):
        return len(self.data_dirs) * self.slice_count

    def __getitem__(self, index):
        path = self.data_dirs[index // self.slice_count]
        image = MetadataSample(path, self.target_shape, gt_type=self.gt_type)
        i = index % self.slice_count + 3
        img_slice = Tensor(image.image[i])
        image_slice = torch.stack([img_slice] * 3)
        gt = image.gt
        if self.transform is not None:
            image_slice = self.transform.single_image(image_slice)
        return image_slice, gt

    def plot(self, index=None):
        if index is None:
            index = np.random.randint(len(self))

        image, gt = self[index]
        image = image.numpy()
        gt = gt.numpy()
        if len(image.shape) == 4:
            image = image[9]
            gt = image[9]
        image_gt = image.copy()
        image_gt[gt != 0] = gt[gt != 0] / gt.max() * image.max()
        for i, img in enumerate([image, gt, image_gt]):
            plt.subplot(1, 3, i + 1)
            plt.imshow(img.T[::-1])
            plt.xticks([])
            plt.yticks([])
        shape = image.shape
        plt.tight_layout(pad=0.2)
        plt.title(f"Image ({index=}) ({shape=}) ")
        plt.show()

    def plot_augmentations(self, index=None, num_images=16, add_gt=False):
        if index is None:
            index = np.random.randint(len(self))

        for i in range(num_images):
            image, gt = self[index]
            image = image.numpy()
            if add_gt:
                image[gt != 0] = gt[gt != 0] / gt.max() * image.max()

            plt.subplot(2, num_images // 2, i + 1)
            plt.imshow(image.T[::-1])
            plt.xticks([])
            plt.yticks([])
        plt.tight_layout(pad=0)
        # plt.title(f"Image ({index=}) ({image.shape=}) ")
        plt.gcf().set_dpi(300)
        plt.show()

    def plot_gt_variations(self, index=9):
        if index is None:
            index = np.random.randint(len(self))
        previous_gt_format = self.gt_format

        from dataclasses import fields
        from itertools import product
        from typing import get_args

        field_values = [get_args(field.type) for field in fields(GTFormat)]
        all_possible_args = list(product(*field_values))

        fig, axes = plt.subplots(2, len(all_possible_args) // 2, figsize=(10, 5))

        for i, (ax, values) in enumerate(zip(axes.flatten(), all_possible_args), 1):
            self.gt_format = GTFormat(*values)

            start_time = time.time()
            _, gt = self[index]
            print(f"{i}. values={values} elapsed_time={time.time() - start_time}")

            gt = gt.numpy()

            ax.imshow(gt.T[::-1])
            ax.set_title("+".join(values), fontsize=6)
            ax.set_xticks([])
            ax.set_yticks([])
        plt.tight_layout(pad=0)
        self.gt_format = previous_gt_format
        # plt.title(f"Image ({index=}) ({image.shape=}) ")
        plt.gcf().set_dpi(300)
        plt.show()


class MetadataPredictionModule(pl.LightningDataModule):
    def __init__(self, data_dir=NAKO_DATASET_PATH, batch_size=1, num_workers=2, gt_type: MetadataTypes = "weight"):
        super().__init__()
        self.batch_size = batch_size
        gt_available = set(get_pid_to_index_lookup().keys())
        self.data_dirs = get_potential_data_dirs(data_dir, required_suffixes=[""], intersection_with=gt_available)
        train_size = int(len(self.data_dirs) * 0.9)
        train_paths = self.data_dirs[:train_size]
        val_paths = self.data_dirs[train_size:]

        self.dataloader_kwargs = dict(
            batch_size=batch_size,
            num_workers=num_workers,
        )

        dataset_kwargs = dict(
            gt_type=gt_type,
        )

        self.train_dataset = MetadataPredictionDataset(train_paths, **dataset_kwargs)
        self.val_dataset = MetadataPredictionDataset(val_paths, **dataset_kwargs)

    def train_dataloader(self):
        return DataLoader(self.train_dataset, shuffle=True, **self.dataloader_kwargs)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, **self.dataloader_kwargs)


def mean_prediction():
    # datamodule = MetadataPredictionModule()
    # train = datamodule.train_dataset
    # gts = np.array([gt for _, gt in train])
    # val = datamodule.val_dataset

    from spine_segmentation.resources.preloaded import get_measure_statistics

    stats = get_measure_statistics()

    gt_types = ["age", "sex", "weight", "size"]
    gts = []
    for i in range(len(stats.dicom_metadata)):
        row = stats.dicom_metadata.iloc[i]
        if i % 100 == 0:
            print(i)
        gt = []

        def handle_gt(gt_type):
            value = row[f"DICOM;Patient{gt_type.capitalize()}"]
            if gt_type == "sex":
                assert value in "MF"
                gt.append((value == "M"))
            elif gt_type == "age":
                assert value.endswith("Y")
                gt.append(int(value[:-1]) / 100)
            elif gt_type == "weight":
                gt.append(float(value) / 100)
            elif gt_type == "size":
                gt.append(float(value))
            else:
                raise ValueError(f"Unknown gt_type {gt_type=} {value=}")

        if isinstance(gt_types, str):
            gt_types = [gt_types]
        for type_ in sorted(set(gt_types)):
            handle_gt(type_)
        gts.append(gt)
    arr = np.array(gts)
    mean = arr.mean(axis=0)
    mse = ((arr - mean) ** 2).mean(axis=0)
    msvariance = mse**0.5
    print(arr.shape)


if __name__ == "__main__":
    # noinspection PyUnreachableCode
    if True:
        mean_prediction()



"""Module to generate .obj file from bronchus coords outer shell

Needs stage-02 (reduced_model.npz) as input.

This file expects the first argument to be a path to the input folder (not including the
reduced_model.npz). The second argument needs to be a path to the output file.

It does so by going through every bronchus point in the reduced_model.npz and checking
each of the 6 neighboring points whether it is empty. If it is indeed empty then
it adds the points which haven't been added yet on that face and also adds the face.

Then it saves everything into a .obj file with the format of [patient_it].obj. This file can
be imported into blender, 3D printed, visualized and many other nice things.
"""

from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Set, Union

import numpy as np

from spine_segmentation.visualisation.color import Color

# from spine.spine_types.arraytype import Vol_3D_I, Mat_4x4_F, List_Vec_3D_F, Vec_3D_F
# from spine.spine_types.color import Color
# from spine.util.geom_types import Ellipsoid


def generate_obj(
    output_data_path: Path,
    model,
    accepted_types: Set[int] = {},
    color_mask=None,
    type_to_color: Union[Color, Dict[int, Color]] = {},
    rot_mat=None,
    num_decimal_digits: int = 4,
):
    """Saves a .obj obj_file given the model, the accepted types and a name

    output_data_path is a pathlib Path, this is the full path the obj_file will be saved as

    model is the 3D numpy array model of the lung or whatever object you
    want to convert to 3D

    accepted_types is a list or set which types should be looked for, other
    types will be ignored. If empty set then everything except for 0 will be
    accepted

    color_to_rgb_tuple is a Dict which maps the color id used in color mask to a tuple
    of rgb values. This color will be used to color the object with that color.
    (e.g. {1: (255, 120, 150)})

    color_mask is a model with the same shape as model, but its numbers represent
    groups of colors/materials which will be added by this script

    rot_mat is a rotation matrix. Each point p=(x,y,z) is rotated by rot_mat @ p or left unchanged if None
    """
    # color_to_rgb_tuple = {}
    # for color_id, color in type_to_color.items():
    #     color_to_rgb_tuple[color_id] = color.floats()[:3]

    output_data_path = Path(output_data_path)

    # Add padding
    model = np.pad(model, 1, mode="constant", constant_values=0)
    if color_mask is not None:
        color_mask = np.pad(color_mask, 1, mode="constant", constant_values=0)

    def color_from_type(type_):
        if isinstance(type_to_color, Color):
            col = type_to_color
        else:
            col = type_to_color.get(material, Color.random())
        return col.floats()[:3]

    vertices = {}
    faces: Dict[int, List[List[int]]] = defaultdict(list)

    reference_shape = model.shape
    model = model.copy()
    if accepted_types:
        for remove in set(np.unique(model)) - accepted_types - {0}:
            model[model == remove] = 0

    index = 1
    # Iterate over each axis and pos/neg directions, then roll the model over, afterwards subtracting these.
    # This causes there to be only -1 and 1 values where there is air, meaning there a face should be added.
    # Though we only look at the 1 values, since these are actually in the model (the others are outside, which means
    # their color map will be wrong)
    for axis in range(3):
        for pos_or_neg in [-1, 1]:
            diff = np.roll(model, -pos_or_neg, axis=axis)
            model_diff = np.where(model > diff)
            coords = []
            for d1, d2 in [(0.5, 0.5), (-0.5, 0.5), (-0.5, -0.5), (0.5, -0.5)]:
                coords.append(list(map(lambda t: t.astype(float), np.copy(model_diff))))
                coords[-1][axis] += pos_or_neg / 2
                coords[-1][(axis + 1) % 3] += d1
                coords[-1][(axis + 2) % 3] += d2

            # Example shape after transpose: (11523, 4, 3) - face_coords is an array of faces,
            # each face is a list of 4 points with 3 coordinates.
            faces_coords = np.transpose(coords, axes=(2, 0, 1))
            vertex_coords = np.transpose(model_diff)
            for vertex_coord, face_coords in zip(vertex_coords, faces_coords):
                curr_face = []
                for face_coord in map(tuple, face_coords):
                    if face_coord not in vertices:
                        vertices[face_coord] = index
                        index += 1
                    curr_face.append(vertices[face_coord])
                material = color_mask[tuple(vertex_coord)] if color_mask is not None else 0
                faces[material].append(curr_face)

    # make to numpy for easier usage later
    vertices = np.array([np.array(v) for v in vertices])
    vertices = normalize(vertices, rot_mat=rot_mat)

    # Write vertices and faces to obj_file
    material_path = output_data_path.with_suffix(".mtl")
    with open(material_path, "w") as mat_file:
        for material in faces:
            mat_file.write(f"newmtl mat{material}\n")
            mat_file.write("Ns 96.078431\n")
            mat_file.write("Ka 1.000000 1.000000 1.000000\n")
            rgb = color_from_type(material)
            mat_file.write(f"Kd {' '.join(map(str, rgb))}\n")
            mat_file.write("Ks 0.500000 0.500000 0.500000\n")
            mat_file.write("Ke 0.000000 0.000000 0.000000\n")
            mat_file.write("Ni 1.000000\n")
            mat_file.write("d 1.000000\n")
            mat_file.write("illum 2\n\n")
    with open(output_data_path, "w") as obj_file:
        obj_file.write("# .obj generated by Airway")
        obj_file.write(f"mtllib {material_path.name}\n")
        obj_file.write("# Vertices\n")
        # original was [[0, 0, -1], [-1, 0, 0], [0, 1, 0]]
        vertexformat = f"v {{:.{num_decimal_digits}f}} {{:.{num_decimal_digits}f}} {{:.{num_decimal_digits}f}}\n"
        for x, y, z in vertices:
            obj_file.write(vertexformat.format(x, y, z))

        obj_file.write("\n# Faces\n")
        for material, faces_with_material in faces.items():
            obj_file.write(f"usemtl mat{material}\n")
            for a, b, c, d in faces_with_material:
                obj_file.write(f"f {a} {b} {c} {d}\n")


def normalize(vertices, reference_shape: np.ndarray = (18, 320, 896), rot_mat=None):
    if reference_shape is not None:
        # Shift to middle of the space
        vertices -= np.array(reference_shape) / 2
        # Scale to [-10..10]
        vertices *= 20 / np.max(reference_shape)
    # If available: transform
    # Note: since this is applied afterwards, points can be out of [-10..10]
    if rot_mat is not None:
        # Add 4th dimension as 1 for transformation matrix
        ones_vec3_to_vec4 = np.ones((len(vertices), 1))
        vertices = np.append(vertices, ones_vec3_to_vec4, axis=1)
        # Transform
        vertices = vertices @ np.transpose(rot_mat)
        # Remove 4th dimension from vectors
        vertices = vertices[:, :-1]
    return vertices


# def generate_ellipsoid_obj(
#         filepath: Path,
#         ellipsoid: Ellipsoid,
#         reference_shape: Vec_3D_F,
#         rot_mat: Mat_4x4_F = None,
#         n_latitude: int = 10,
#         n_longitude: int = 10,
# ):
#     with open(filepath, "w") as file:
#         center, radii = ellipsoid.center, ellipsoid.radii
#         vertices = [center + radii * [0, 0, 1], center - radii * [0, 0, 1]]  # Top and bottom vertices
#         curr_vertex_index = 3  # 1 and 2 are the above points (.obj files are 1-indexed)
#         latitude_point_indices = []  # Fill with lists of indexes for each latitude, used to connect faces later
#
#         for lon_angle in np.linspace(0, 2 * np.pi, n_longitude, endpoint=False):  # Traverse angle along longitude
#             latitude_point_indices.append(list(range(curr_vertex_index, curr_vertex_index + n_latitude - 2)))
#             curr_vertex_index += n_latitude - 2
#             for lat_angle in np.linspace(0, np.pi, n_latitude)[1:-1]:  # Skip existing first and last points
#                 polar_angle = np.array(
#                     [np.sin(lat_angle) * np.cos(lon_angle), np.sin(lat_angle) * np.sin(lon_angle), np.cos(lat_angle)]
#                 )
#                 vertices.append(center + radii * polar_angle)
#         # Normalize twice, once with inverse, then translate with reference shape, then retransform to align coordinates
#         for vertex in normalize(normalize(vertices, None, np.linalg.inv(rot_mat)), reference_shape, rot_mat):
#             file.write(f"v {' '.join(map(str, vertex))}\n")
#         for left, right in zip(latitude_point_indices, latitude_point_indices[1:] + latitude_point_indices[:1]):
#             file.write(f"f 1 {left[0]} {right[0]}\n")  # First triangle to top point
#             file.write(f"f 2 {left[-1]} {right[-1]}\n")  # Last triangle to bottom point
#             for left2points, right2points in zip(zip(left, left[1:]), zip(right, right[1:])):
#                 file.write(f"f {' '.join(map(str, left2points + right2points[::-1]))}\n")  # Connect squares



import sys

from pytorch_lightning.callbacks import TQDMProgressBar


class FixValidationStepBar(TQDMProgressBar):
    # https://github.com/Lightning-AI/lightning/issues/15283#issuecomment-1289654353
    def init_validation_tqdm(self):
        bar = super().init_validation_tqdm()
        if not sys.stdout.isatty():
            bar.disable = True
        return bar

    def init_predict_tqdm(self):
        bar = super().init_predict_tqdm()
        if not sys.stdout.isatty():
            bar.disable = True
        return bar

    def init_test_tqdm(self):
        bar = super().init_test_tqdm()
        if not sys.stdout.isatty():
            bar.disable = True
        return bar



from spine_segmentation.resources.preloaded import get_measure_statistics


def stat(df):
    import matplotlib.pyplot as plt

    # Calculate the BMI using the DICOM;PatientSize (meters) and DICOM;PatientWeight (kg) columns
    df["BMI"] = df["DICOM;PatientWeight"] / (df["DICOM;PatientSize"] ** 2)

    # Extract the Volume data from the C2-3;DISC_VOLUME column
    volume = df["T10-11;DISC_VOLUME"]
    # df.to_html("index.html")
    # print(df.corr())

    # Plot the BMI over Volume
    plt.scatter(df["BMI"], volume, s=1)
    # plt.ylabel('Volume (mm^3)')
    plt.ylabel("Volume (mm^3)")
    plt.xlabel("BMI")
    plt.title("Volume over BMI")
    plt.show()


stat(get_measure_statistics().raw_statistics)



import time
from functools import lru_cache
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Union

import lightning.pytorch as pl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from spine_segmentation.datasets.augmentation import ImageAugmentation
from spine_segmentation.datasets.path_helper import expand_path_to_data_dirs
from spine_segmentation.datasets.sample import GTFormat, Sample
from spine_segmentation.resources.paths import NAKO_DATASET_PATH
from spine_segmentation.resources.other_paths import RAW_NAKO_DATASET_PATH


class SegmentationDataset(Dataset):
    def __init__(
        self,
        data_dirs: Iterable[Path],
        transform=None,
        *,
        gt_format: Optional[GTFormat] = GTFormat(),
        target_shape: Tuple[Optional[int], int, int] = (18, 320, 896),
        slice_wise: bool = True,
        add_bs_wk_as_channels: bool = False,
        add_adjacent_slices: bool = False,
        triple_same_slice: bool = False,
        crop_height_to_px: int = None,
        use_only_n_center_slices: Optional[int] = None,
        overwrite_length: int = None,
        index_to_zip_path: Dict[int, Path] = None,
        index_to_slice_index: Dict[int, int] = None,
        zip_to_gt_path_lookup: Dict[Path, Path] = None,
    ):
        super().__init__()
        self.target_shape = target_shape
        self.data_dirs = list(data_dirs)
        self.gt_format = gt_format
        self.slice_wise = slice_wise
        self.transform = transform
        self.add_bs_wk_as_channels = add_bs_wk_as_channels
        self.add_adjacent_slices = add_adjacent_slices
        self.triple_same_slice = triple_same_slice
        self.crop_height_to_px = crop_height_to_px or self.target_shape[2]
        self.use_only_n_center_slices = use_only_n_center_slices or self.target_shape[0]
        self.overwrite_length = overwrite_length
        self.zip_to_gt_path_lookup = zip_to_gt_path_lookup
        self.index_to_zip_path = index_to_zip_path
        self.index_to_slice_index = index_to_slice_index

        assert (
            add_adjacent_slices + triple_same_slice + add_bs_wk_as_channels <= 1
        ), "Only one of these can be true: add_adjacent_slices, triple_same_slice, add_bs_wk_as_channels"
        if self.use_only_n_center_slices is None:
            self.slice_count = None
        else:
            self.slice_count = (self.use_only_n_center_slices - self.add_adjacent_slices * 2) if self.slice_wise else 1

    def __len__(self):
        if self.overwrite_length is not None:
            return self.overwrite_length
        return len(self.data_dirs) * self.slice_count

    @lru_cache(maxsize=5)
    def _get_sample(self, path):
        return Sample(
            path,
            self.target_shape,
            gt_format=self.gt_format,
            crop_height_to_px=self.crop_height_to_px,
            zip_to_gt_path_lookup=self.zip_to_gt_path_lookup,
        )

    def __getitem__(self, index):
        if self.index_to_zip_path is not None:
            path = self.index_to_zip_path[index - self.index_to_slice_index[index]]
        else:
            path = self.data_dirs[index // self.slice_count]
        sample = self._get_sample(path)

        # print(index, sample, sample.image.shape, sample.gt.shape)

        if self.use_only_n_center_slices is None:
            start_at = 0
        else:
            start_at = self.target_shape[0] // 2 - self.use_only_n_center_slices // 2

        if self.slice_wise:
            if self.index_to_slice_index is not None:
                i = self.index_to_slice_index[index]
            else:
                i = start_at + index % self.slice_count
            # return sample.sample[i : i + 1], sample.gt_onehot[:, i : i + 1]
            if self.add_adjacent_slices:
                # image_slice = Tensor(sample.image[i : i + 3])
                curr_slice = Tensor(sample.image[i])
                valid_indices = range(sample.image.shape[0])
                image_slice = torch.stack(
                    [
                        Tensor(sample.image[i - 1]) if i - 1 in valid_indices else curr_slice,
                        curr_slice,
                        Tensor(sample.image[i + 1]) if i + 1 in valid_indices else curr_slice,
                    ],
                    dim=0,
                )
                gt_slice = Tensor(sample.gt[i : i + 1].astype(np.int16))
            else:
                image_slice, gt_slice = Tensor(sample.image[i : i + 1]), Tensor(sample.gt[i : i + 1].astype(np.int16))

            if self.add_bs_wk_as_channels:
                image_slice = torch.cat(
                    [
                        image_slice,
                        Tensor(sample.wk[i : i + 1]),
                        Tensor(sample.bs[i : i + 1]),
                    ],
                    dim=0,
                )
            if self.triple_same_slice:
                image_slice = torch.cat([image_slice, image_slice, image_slice], dim=0)
            if self.transform is not None:
                image_slice, gt_slice = self.transform.single_image(image_slice, gt_slice)
            # assert image_slice.shape == gt_slice.shape, f"{image_slice.shape=} != {gt_slice.shape=}"
            return image_slice, gt_slice
        else:
            assert not any((self.add_adjacent_slices, self.triple_same_slice, self.add_bs_wk_as_channels))
            return sample.image, sample.gt

    def plot(self, index=None):
        if index is None:
            index = np.random.randint(len(self))

        image, gt = self[index]
        image = image.numpy()
        gt = gt.numpy()
        if len(image.shape) == 4:
            image = image[9]
            gt = image[9]
        image_gt = image.copy()
        image_gt[gt != 0] = gt[gt != 0] / gt.max() * image.max()
        for i, img in enumerate([image, gt, image_gt]):
            plt.subplot(1, 3, i + 1)
            plt.imshow(img.T[::-1])
            plt.xticks([])
            plt.yticks([])
        shape = image.shape
        plt.tight_layout(pad=0.2)
        plt.title(f"Image ({index=}) ({shape=}) ")
        plt.show()

    def plot_augmentations(self, index=None, num_images=16, add_gt=False):
        if index is None:
            index = np.random.randint(len(self))

        for i in range(num_images):
            image, gt = self[index]
            image = image.numpy()
            if add_gt:
                image[gt != 0] = gt[gt != 0] / gt.max() * image.max()

            plt.subplot(2, num_images // 2, i + 1)
            plt.imshow(image.T[::-1])
            plt.xticks([])
            plt.yticks([])
        plt.tight_layout(pad=0)
        # plt.title(f"Image ({index=}) ({image.shape=}) ")
        plt.gcf().set_dpi(300)
        plt.show()

    def plot_gt_variations(self, index=9):
        if index is None:
            index = np.random.randint(len(self))
        previous_gt_format = self.gt_format

        from dataclasses import fields
        from itertools import product
        from typing import get_args

        field_values = [get_args(field.type) for field in fields(GTFormat)]
        all_possible_args = list(product(*field_values))

        fig, axes = plt.subplots(2, len(all_possible_args) // 2, figsize=(10, 5))

        for i, (ax, values) in enumerate(zip(axes.flatten(), all_possible_args), 1):
            self.gt_format = GTFormat(*values)

            start_time = time.time()
            _, gt = self[index]
            print(f"{i}. values={values} elapsed_time={time.time() - start_time}")

            gt = gt.numpy()

            ax.imshow(gt.T[::-1])
            ax.set_title("+".join(values), fontsize=6)
            ax.set_xticks([])
            ax.set_yticks([])
        plt.tight_layout(pad=0)
        self.gt_format = previous_gt_format
        # plt.title(f"Image ({index=}) ({image.shape=}) ")
        plt.gcf().set_dpi(300)
        plt.show()


class SegmentationDataModule(pl.LightningDataModule):
    def __init__(
        self,
        data_dir: Union[str, Path, List[str], Tuple[str]] = NAKO_DATASET_PATH,
        batch_size=1,
        num_workers=2,
        gt_format: GTFormat = GTFormat(),
        target_shape: Tuple[Optional[int], int, int] = (18, 320, 896),
        slice_wise=True,
        augment=False,
        add_bs_wk_as_channels=False,
        add_adjacent_slices=False,
        triple_same_slice=False,
        crop_height_to_px: int = None,
        use_only_n_center_slices: int = None,
        ignore_train_from_csv: List[Union[Path, str]] = None,
    ):
        super().__init__()
        self.batch_size = batch_size

        train_index_length = None
        val_index_length = None

        train_index_to_zip_path = None
        val_index_to_zip_path = None

        index_to_slice_index = None
        val_index_to_slice_index = None

        index_to_npz_path = None

        zip_path_to_npz_path = None

        if isinstance(data_dir, tuple) or isinstance(data_dir, list):
            train_path, val_path = data_dir
            train_paths = expand_path_to_data_dirs(train_path)
            val_paths = expand_path_to_data_dirs(val_path)
        else:
            self.data_dirs = expand_path_to_data_dirs(data_dir)

            assert self.data_dirs, f"No data found in {data_dir=}"

            train_size_percentage = 0.95
            train_size = int(len(self.data_dirs) * train_size_percentage)
            train_paths = self.data_dirs[:train_size]
            val_paths = self.data_dirs[train_size:]

            if Path(data_dir).suffix == ".npz":
                npz = np.load(data_dir, allow_pickle=True)
                index_to_zip_path = npz["index_to_zip_path"].item()
                index_to_slice_index = npz["index_to_slice_index"].item()
                index_to_npz_path = npz["index_to_npz_path"].item()

                common_indices = (
                    set(index_to_npz_path.keys()) & set(index_to_zip_path.keys()) & set(index_to_slice_index.keys())
                )
                index_to_zip_path = {index: index_to_zip_path[index] for index in common_indices}
                assert all(index in index_to_slice_index for index in common_indices)
                index_to_npz_path = {index: index_to_npz_path[index] for index in common_indices}
                assert all(index in index_to_zip_path for index in common_indices)
                assert all(index in index_to_npz_path for index in common_indices)

                first_val_index = next(index for index, path in index_to_zip_path.items() if path == val_paths[0])
                train_index_length = first_val_index - 1
                val_index_length = len(index_to_slice_index) - train_index_length
                assert val_index_length + train_index_length == max(index_to_slice_index) + 1
                val_index_length -= 21

                train_index_to_zip_path = {
                    index: path for index, path in index_to_zip_path.items() if index < first_val_index
                }
                val_index_to_zip_path = {
                    index - first_val_index: path
                    for index, path in index_to_zip_path.items()
                    if index >= first_val_index
                }

                val_index_to_slice_index = {
                    index - first_val_index: slice_index
                    for index, slice_index in index_to_slice_index.items()
                    if index >= first_val_index
                }

                common_indices = set(index_to_npz_path.keys()) & set(index_to_zip_path.keys())
                zip_path_to_npz_path = {index_to_zip_path[index]: index_to_npz_path[index] for index in common_indices}

        if ignore_train_from_csv is not None:
            for ignore_path in ignore_train_from_csv:
                ignore_csv = pd.read_csv(ignore_path)
                ignore_paths = ignore_csv["Path"]

                print("============== Ignore ==============")
                print(f"{len(train_paths)=}")
                print(f"{len(ignore_paths)=}")
                ignore_paths = {Path(path).name for path in ignore_paths}
                print(list(ignore_paths)[:10])
                print("============")
                print(train_paths[:10])
                print("============")
                train_paths = [
                    p for p in train_paths if p.name not in ignore_paths and p.name.split("_")[0] not in ignore_paths
                ]
                print(f"{len(train_paths)=}")

        self.segmentation_dataloader_kwargs = dict(
            batch_size=batch_size,
            num_workers=num_workers,
        )

        segmentation_dataset_kwargs = dict(
            gt_format=gt_format,
            slice_wise=slice_wise,
            add_bs_wk_as_channels=add_bs_wk_as_channels,
            add_adjacent_slices=add_adjacent_slices,
            triple_same_slice=triple_same_slice,
            crop_height_to_px=crop_height_to_px,
            use_only_n_center_slices=use_only_n_center_slices,
            zip_to_gt_path_lookup=zip_path_to_npz_path,
            target_shape=target_shape,
        )

        train_args = dict(
            transform=ImageAugmentation() if augment else None,
            overwrite_length=train_index_length,
            index_to_zip_path=train_index_to_zip_path,
            index_to_slice_index=index_to_slice_index,
        )

        val_args = dict(
            overwrite_length=val_index_length,
            index_to_zip_path=val_index_to_zip_path,
            index_to_slice_index=val_index_to_slice_index,
        )

        self.train_dataset = SegmentationDataset(train_paths, **train_args, **segmentation_dataset_kwargs)
        self.val_dataset = SegmentationDataset(val_paths, **val_args, **segmentation_dataset_kwargs)
        from spine_segmentation.utils.globals import TRAIN_DATASET_SIZE

        TRAIN_DATASET_SIZE.set(len(self.train_dataloader()))

    def train_dataloader(self):
        return DataLoader(self.train_dataset, **self.segmentation_dataloader_kwargs)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, **self.segmentation_dataloader_kwargs)


if __name__ == "__main__":
    # noinspection PyUnreachableCode
    if False:
        dataset = SegmentationDataset(get_potential_data_dirs(NAKO_DATASET_PATH))
        dataset.plot()

    # noinspection PyUnreachableCode
    if False:
        dataset = SegmentationDataset(get_potential_data_dirs(NAKO_DATASET_PATH), ImageAugmentation())
        dataset.plot_augmentations(add_gt=True)

    # noinspection PyUnreachableCode
    if False:
        dataset = SegmentationDataset(get_potential_data_dirs(NAKO_DATASET_PATH))
        dataset.plot_gt_variations()

    # noinspection PyUnreachableCode
    if True:
        dataset = SegmentationDataset(list(RAW_NAKO_DATASET_PATH.glob("*.zip"))[:100], gt_format=None)
        print(dataset[0])



from pathlib import Path

from spine_segmentation.datasets.sample import SampleIterator
from spine_segmentation.inference.onnx_model import ONNXInferenceModel
from spine_segmentation.plotting.plot_slice import plot_npz
from spine_segmentation.utils.log_dir import get_next_log_dir

prefix = Path("define path")

edge_case_paths = [
    "1b7cd7a47cc95032792b48fd5b6f5138.npz",
    "3f776bfd85beb75318e5e2f51dcc1cef.npz",
    "68958c058fc83017a3d8e5e00a9321fb.npz",
    "8897172fff245644f0d3acf7552f090d.npz",
    "00cccb4bb3680540355ea2dea7c7a936.npz",
    "0eba72dd557eb2397e3b3cde07166cb3.npz",
    "b4d9152a99be88e7fbfde19547b42e3e.npz",
    "0ee5499747f9332f9fa4036eae43e2f3.npz",
    "9080e06c4c8601213098acaccee44e46.npz",
    "5781cb28b20d9fc416aec7b04b842e07.npz",
    "5530f3e564d5ae1ec2508dce02fef34f.npz",
    "74e0c90952a4a3022b188a1b0ed28ceb.npz",
    "6ede6800b5ab8915fea0ffde354a662b.npz",
    "b2ec9888dce9c6eb145cd3f3666ffda7.npz",
    "ade5f227c3da97d63fcb40ddd2fc3e30.npz",
    "22bf344080942957ae839dc344377658.npz",
    "c0fd29f1b15373ecd5edd2d5c5817855.npz",
    "76e1ab95dab4d0ec002211a1bfbb458b.npz",
    "d329cc194c630ced2401ec1baafee0fc.npz",
    "a509f7e5ca2595ff20351689069209ee.npz",
    "40a706777ec5274ed83ffa73d65e0ff8.npz",
    "7b409419cced71aa135f429afb908db0.npz",
    "a461064d250933289c43f37ddfa28696.npz",
    "85ddb7690a26ba0e416fa315c65b14f8.npz",
    "4644386d37e5d829f7ea2e8d5f24f5b0.npz",
    "c8248089d8ae38bfa8f62d215afffc8b.npz",
    "038e045c136ff888060da7ec3bacc301.npz",
    "481794b8cc08d88cdaeb076998cf39b1.npz",
    "7fe1fa781c990aaf70e9e9ee2ebbd911.npz",
    "5f3a78ed8e015e8422503b29f15519e8.npz",
    "778dbc632584c4039576b64ee9812553.npz",
    "3280b390a96b2616b71e9db3ed828b53.npz",
    "6f18fb717f873c42b326ad8fac7c971f.npz",
    "3d8f8d80b0d95d5ba432866856ab6c70.npz",
    "869d8379b5ddd582043f6fa845b878a7.npz",
]


def main():
    onnx = ONNXInferenceModel.get_best_segmentation_model()
    onnx.load_index_list()
    npz_path_to_zip_path = {npz_path: onnx.index_to_zip_path[i] for i, npz_path in onnx.index_to_npz_path.items()}
    zip_paths = [npz_path_to_zip_path[prefix / path] for path in edge_case_paths]
    sample_iterator = SampleIterator(zip_paths, add_adjacent_slices=True)

    log_dir = get_next_log_dir()

    for image, gt, sample, path in sample_iterator:
        print(image.shape, gt.shape)
        npz = onnx.inference(image, gt=gt)
        save_to = log_dir / "edge_cases" / f"{sample.dicom.PatientID}.png"
        plot_npz(npz, save_to, slices=range(0, 100))


if __name__ == "__main__":
    main()



from ydata_profiling import ProfileReport

from spine_segmentation.resources.preloaded import get_measure_statistics

if __name__ == "__main__":
    profile_report = ProfileReport(get_measure_statistics().dicom_metadata)
    profile_report.to_file("profiling.html")



from collections import Counter
from pathlib import Path

import numpy as np
import segmentation_models_pytorch as smp
import torch
from matplotlib import pyplot as plt

from spine_segmentation.instance_separation.instance_separation import separate_rois_with_labels
from spine_segmentation.spine_types.labels import get_labels_for_n_classes
from spine_segmentation.visualisation.color_palettes import get_alternating_palette, get_wk_bs_palette


def is_completely_correct(prediction, gt):
    correct_no_edges = True
    ids_correct = Counter()
    for curr, count in zip(*np.unique(gt, return_counts=True)):
        if curr == 0:
            continue
        if count < 20:
            continue

        gt_mask = gt == curr
        occurrence = np.bincount(prediction[gt_mask])
        occurrence[0] = 0
        second_most_common = occurrence.argmax()
        if second_most_common != curr:
            correct_no_edges = False
        ids_correct[curr] += second_most_common == curr
    return correct_no_edges, ids_correct


def get_stats(prediction, gt):
    print(np.unique(prediction), np.unique(gt))
    prediction = torch.tensor(prediction.copy())
    gt = torch.tensor(gt.copy())

    remove_px = 15
    zero_layers = sum(torch.amin(prediction == 0, dim=(0, 1)))
    top_remove_px = remove_px + zero_layers
    print(f"{remove_px=}, {top_remove_px=}")

    appeared_counter = Counter(np.unique(gt))

    no_edges_gt = gt[:, :, remove_px:-top_remove_px]
    appeared_no_edges_counter = Counter(np.unique(no_edges_gt))

    correct, correct_counter = is_completely_correct(prediction, gt)
    correct_no_edges, correct_no_edges_counter = is_completely_correct(
        prediction[:, :, remove_px:-top_remove_px], no_edges_gt
    )

    print(f"{correct=}, {correct_no_edges=}")
    # prediction = torch.nn.functional.one_hot(prediction, num_classes=50)[..., 1:]
    # gt = torch.nn.functional.one_hot(gt, num_classes=50)[..., 1:]

    # tp, fp, fn, tn = smp.metrics.get_stats(prediction, gt, mode="multilabel")
    tp, fp, fn, tn = smp.metrics.get_stats(prediction - 1, gt - 1, mode="multiclass", num_classes=50, ignore_index=-1)
    kwargs = dict(tp=tp, fp=fp, fn=fn, tn=tn, reduction="micro")
    stats = dict(
        # prediction=prediction,
        # gt=gt,
        tp=tp,
        fp=fp,
        fn=fn,
        tn=tn,
        correct=correct,
        correct_counter=correct_counter,
        correct_no_edges=correct_no_edges,
        correct_no_edges_counter=correct_no_edges_counter,
        appeared_counter=appeared_counter,
        appeared_no_edges_counter=appeared_no_edges_counter,
        iou=smp.metrics.iou_score(**kwargs),
        f1=smp.metrics.f1_score(**kwargs),
        f2=smp.metrics.fbeta_score(**kwargs, beta=2),
        acc=smp.metrics.accuracy(**kwargs),
        recall=smp.metrics.recall(**kwargs),
        precision=smp.metrics.precision(**kwargs),
    )
    for key, value in stats.items():
        if isinstance(value, torch.Tensor):
            stats[key] = value.numpy()
    return stats


def single_plot(ax, palette, background, instance, labels, title, *, font_scale=1):
    # plt.subplot(1, len(plots), index + 1)

    # colors = plt.cm.prism(np.linspace(0, 1, 50))
    colors = np.array([list(c.floats()) for c in palette.values()])
    colors[0] = 0

    # if background is not None:
    #    colors[:, -1] = 0.1
    #    colors[0, -1] = 0
    if title and "correct" in title.lower():
        colors = [[0, 0, 0, 1], [0, 1, 0, 1], [1, 0, 0, 1]]

    vmax = len(colors) - 1
    cmap = plt.cm.colors.ListedColormap(colors)

    ax.set_xticks([])
    ax.set_yticks([])

    if title is not None:
        ax.set_title(title, fontsize=8 * font_scale)

    if labels is not None:
        for i in np.unique(instance)[1:]:
            centroid = np.median(np.nonzero(instance == i), axis=1)
            label = labels.get(i, str(i))
            is_disc = "-" in label
            if not is_disc:
                ax.text(
                    centroid[1] + (45 if not is_disc else 0),
                    centroid[0],
                    label,
                    ha="left" if not is_disc else "center",
                    va="center",
                    fontsize=4 * font_scale,
                    color="black",
                    # path_effects=[patheffects.withStroke(linewidth=0.5, foreground="black")],
                )
                ax.plot([centroid[1], centroid[1] + 43], [centroid[0], centroid[0]], color="black", linewidth=0.3)

    if background is not None:
        ax.imshow(background)
    if instance is not None:
        ax.imshow(instance, cmap=cmap, vmin=0, vmax=vmax, interpolation="none")


def plot_npz(npz, save_to: str = None, show: bool = False, slices: range = None, save_separate=False, stats=None):
    get_transposed = lambda key: npz[key].T[::-1]

    if "gt_instances" in npz:
        gt_instances = npz["gt_instances"]
        gt = gt_instances.T[::-1]
    elif list(np.unique(npz["gt"])) == [0]:
        gt_instances = np.zeros_like(npz["segmentation"])
        gt = get_transposed("gt")
    else:
        gt_instances, _ = separate_rois_with_labels(npz["gt"])
        gt = get_transposed("gt")

    gt_instances = gt_instances.T[::-1]

    instances = get_transposed("instances")

    if "instances_post_processed" in npz:
        output = get_transposed("instances_post_processed")
    elif "segmentation" in npz:
        output = get_transposed("segmentation")
    else:
        output = instances

    if stats is None:
        stats = get_stats(npz["instances_post_processed"], npz["gt_instances"])
        # stats = get_stats(get_transposed("instances_post_processed").T, gt_instances.T)

    footer = (
        f"iou={stats['iou']:.3f}  "
        f"f1={stats['f1']:.3f}  "
        f"f2={stats['f2']:.3f}  "
        f"prec={stats['precision']:.3f}  "
        f"rec={stats['recall']:.3f}  "
        f"acc={stats['acc']:.3f}  "
    )

    original = get_transposed("original")[:, :, 0, :]

    if slices is None:
        slices = range(output.shape[2] // 2 - 1, output.shape[2] // 2)

    slices = range(slices.start, min(slices.stop, output.shape[2]), slices.step)

    # id_to_label = npz["id_to_label"].item()

    gt_id_to_label = id_to_label = dict(zip(range(1, 50), get_labels_for_n_classes(49)))

    # Binary dilation
    if use_highlighting := False:
        from scipy.ndimage import binary_dilation

        vertebra = output % 2 == 1
        discs = (output % 2 == 0) & (output != 0)
        highlight_vertebra = binary_dilation(vertebra) & ~vertebra
        highlight_discs = binary_dilation(discs) & ~discs
        highlight_overall = binary_dilation(output != 0) & ~(output != 0)

        highlight = highlight_vertebra | highlight_discs | highlight_overall
        highlighted = original + highlight * original.max() / 2

    alternating = get_alternating_palette()
    wkbs_palette = get_wk_bs_palette()

    plots = [
        [original, None, None, "MRI", alternating],
        # [highlighted, instances, id_to_label, "Orig+Out", alternating],
        # [original, output, None, "Seg", wkbs_palette],
        [None, instances, gt_id_to_label, "Output", alternating],
        # [None, instances, id_to_label, "Output", alternating],
        [None, gt_instances, gt_id_to_label, "GT", alternating],
    ]

    if "instances_post_processed" in npz:
        post_processed = get_transposed("instances_post_processed")
        plots.insert(-1, [None, post_processed, id_to_label, "Post-processed", alternating])

        correct = (post_processed == gt_instances).astype(int)
        correct[post_processed != gt_instances] = 2
        correct[gt_instances == 0] = 0
        title = "Incorrect"
        if stats["correct_no_edges"]:
            title = "Correct (NE)"
        if stats["correct"]:
            title = "Correct"

        incorrect_ids = [
            v for v, k in (stats["appeared_no_edges_counter"] - stats["correct_no_edges_counter"]).items() if k > 0
        ]
        title += f" ({', '.join(map(str, incorrect_ids))})"
        plots.append([None, correct, None, title, alternating])

    if "cropped_segmentation" in npz:
        cropped_segmentation = get_transposed("cropped_segmentation")
        plots.append([None, cropped_segmentation, None, "Segmentation", wkbs_palette])

    # if index == len(plots) // 2:
    thin_by = 30
    thinner = slice(thin_by, original.shape[1] - thin_by)

    # print(original)
    is_not_zero = np.abs(original) > 1e-6
    # print(np.any(is_not_zero, axis=(1, 2)).shape)
    # print(np.any(is_not_zero, axis=(1, 2)))
    zero_layers_at_start = np.argmax(np.any(is_not_zero, axis=(1, 2)))
    zero_layers_at_end = np.argmax(np.any(is_not_zero, axis=(1, 2))[::-1])
    # print(f"{zero_layers_at_start=}, {zero_layers_at_end=}")
    shorter = slice(zero_layers_at_start, original.shape[0] - zero_layers_at_end)

    save_to_ax = save_to.parent / "axes"
    save_to_ax.mkdir(exist_ok=True, parents=True)

    for slice_index in slices:
        fig, axes = plt.subplots(1, len(plots))

        for index, (ax, (background, instance, labels, title, palette)) in enumerate(zip(axes, plots)):
            if background is not None:
                background = background[shorter, thinner, slice_index]
            if instance is not None:
                instance = instance[shorter, thinner, slice_index]
            single_plot(ax, palette, background, instance, labels, title)

            if save_separate:
                single_fig, single_ax = plt.subplots(1, 1)
                single_plot(single_ax, palette, background, instance, labels, title)
                single_fig.savefig(
                    save_to_ax / f"slice{slice_index}_{title.replace(' ', '')}.png", dpi=600, transparent=True
                )
                single_fig.tight_layout(pad=0)
                plt.close(single_fig)

        # text_kwargs = dict(ha="center", va="center", fontdict={"family": "monospace"}, color="black", fontsize=6)
        # for at in np.linspace(0.05, 0.3, 4):
        #     fig.text(0.5, at, footer, **text_kwargs)

        fig.tight_layout(pad=0)

        if save_to is not None:
            save_to = Path(save_to)
            save_to.parent.mkdir(parents=True, exist_ok=True)
            fig.savefig(save_to.with_name(f"{save_to.stem}_slice{slice_index:02}{save_to.suffix}"), dpi=600)
        if show:
            plt.show()
        plt.close()



from pathlib import Path

PROJECT_PATH = Path(__file__).absolute().parent.parent.parent

DATASETS_PATH = PROJECT_PATH / "datasets"
REPORTS_PATH = PROJECT_PATH / "reports"
MODELS_PATH = PROJECT_PATH / "models"
SECRETS_PATH = PROJECT_PATH / "secrets"
LOGS_PATH = PROJECT_PATH / "logs"

PLOTS_PATH = REPORTS_PATH / "plots"
RESULTS_PATH = REPORTS_PATH / "results"

MEASURE_STATISTICS_PATH = DATASETS_PATH / "measure_statistics_new.csv"
VECTOR_TABLE_PATH = DATASETS_PATH / "vector_table_new.npz"

# Paths outside repository

CACHE_PATH = Path("/tmp/spine-segmentation-cache")
CACHE_PATH.mkdir(exist_ok=True)

NAS_RESEARCH_PATH = Path("~/devel/data_remote/nas_research").expanduser().resolve()
NAKO_DATASET_PATH = NAS_RESEARCH_PATH / "path"

LOCAL_NAKO_DATASET_PATH = Path.home() / "Data/nako/2022-08-22"

VAL_SPLIT_CSV_PATH = DATASETS_PATH / "nako_splits/val_seg.csv"
TRAIN_SPLIT_CSV_PATH = DATASETS_PATH / "nako_splits/train_seg.csv"



from collections import defaultdict

import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
from scipy import stats

from spine_segmentation.datasets.metadata_prediction_dataset import MetadataPredictionModule
from spine_segmentation.models.pl_modules.regression_module import RegressionModule
from spine_segmentation.models.regression.regression_model import RegressionModel


def main():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    path = "<chkpt_path>"
    model = RegressionModule.load_from_checkpoint(path, device, model=RegressionModel(4)).eval()
    # model = lightning.pytorch.LightningModule.load_from_checkpoint(path, device).eval()
    # model = torch.load(path, device).eval()

    data_module = MetadataPredictionModule()
    val_data = data_module.val_dataset

    results = []
    for i in range(len(val_data)):
        print(i)
        sample = val_data[i]
        prediction = model.predict(sample.image)
        results.append(prediction)
        results.append(sample.gt)

    df = pd.DataFrame(results)
    df.to_csv("predictions.csv", index=False)


def compare_csv():
    path = "<csv_path>"
    pd.options.display.float_format = "{:.4f}".format
    np.set_printoptions(formatter={"float_kind": pd.options.display.float_format})
    df = pd.read_csv(path)
    df *= 100
    df.iloc[:, 2] /= 100
    df.iloc[:, 6] /= 100
    predicted = df.iloc[:, 1:5].to_numpy()
    correct = df.iloc[:, 5:9].to_numpy()
    plot(predicted, correct)
    # plot(predicted, "Prediction")

    print("Age | Sex | Size | Weight")
    data_variance = calculate_variance(0, correct)
    data_stddev = np.sqrt(data_variance)
    # data_confidence = confidence_interval(0, correct)
    print("Data variance", data_variance)
    print("Data stddev", data_stddev)
    # print("Confidence interval", data_confidence)
    # counts = (data_confidence[0] <= correct) & (correct <= data_confidence[1])
    # print(counts.sum(axis=0))
    # print(predicted)
    # print(correct)
    # diff = (predicted - correct)
    # mse = ((diff**2).mean(axis=0))
    saved = defaultdict(lambda: defaultdict(list))
    sep = "=" * 10
    print(f"{sep} Prediction {sep}")
    evaluate(predicted, correct, saved["EfficientNet"])

    print(f"{sep} Mean predictor {sep}")
    mean = correct.mean(axis=0)
    evaluate(mean, correct, saved["Mean_Predictor"])

    print(f"{sep} Median predictor {sep}")
    median = np.median(correct, axis=0)
    evaluate(median, correct, saved["Median_Predictor"])

    print_for_sheets(saved)


def print_for_sheets(saved):
    print("Copy this for sheets:\n")
    for predictor_name, saved_data in saved.items():
        row = [predictor_name, ""]
        for i in range(4):
            for metric in ["MSE", "Stddev", "MAE"]:
                row.append(str(saved_data[metric][i]))
        print("\t".join(row))


def evaluate(prediction, data, save_in):
    print(prediction)
    print("== MSE")
    print(mse := calculate_mse(prediction, data))
    save_in["MSE"] = mse
    print("== Variance")
    variance = calculate_variance(prediction, data)
    save_in["Variance"] = variance
    stddev = np.sqrt(variance)
    save_in["Stddev"] = stddev
    print(variance)
    print("== Stddev")
    print(stddev)
    print("== MAE")
    print(mae := calculate_mae(prediction, data))
    save_in["MAE"] = mae
    # print("== Confidence Interval 95%")
    # print(confidence_interval(prediction, data))
    print("\n")


def calculate_mse(prediction, data):
    mse = (data - prediction) ** 2
    return mse.mean(axis=0)


def calculate_mae(prediction, data):
    mae = np.abs(data - prediction)
    return mae.mean(axis=0)


def calculate_msd(prediction, data):
    """Mean signed difference"""
    msd = data - prediction
    return msd.mean(axis=0)


def calculate_variance(prediction, data):
    bias = calculate_msd(prediction, data)
    mse = calculate_mse(prediction, data)
    variance = mse - bias**2
    return variance


def confidence_interval(prediction, data, confidence=0.99):
    """Calculate confidence interval for given data and prediction"""
    # Only makes sense for normally distributed data!
    mean = calculate_mae(prediction, data)
    interval = stats.norm.interval(confidence, loc=mean, scale=stats.sem(data))
    return interval


def plot(predicted, data):
    plt.suptitle(f"Data and predictions distribution (n={data.shape[0]})")
    for i, title in zip(range(data.shape[1]), ["Age", "Sex", "Size", "Weight"]):
        plt.subplot(2, 2, i + 1)
        plt.title(title)
        # color = "blue" if title_suffix.lower() == "data" else "red"
        plt.hist(data[:, i], bins=50, color="blue", label="Data")
        plt.hist(predicted[:, i], bins=50, color="red", alpha=0.7, label="Prediction")
    plt.legend()
    plt.show()


if __name__ == "__main__":
    compare_csv()
    # main()



from itertools import zip_longest


def get_vertebra_labels(c_range=range(2, 8), t_range=range(1, 13), l_range=range(1, 6)):
    return [*[f"C{i}" for i in c_range], *[f"T{i}" for i in t_range], *[f"L{i}" for i in l_range], "S1"]


def get_disc_labels(vertebra_labels):
    return [
        *[f"{upper}-{lower}" for upper, lower in zip(vertebra_labels[:-1], vertebra_labels[1:])],
    ]


# 6 + 12 + 5 + 1
def get_labels_for_n_classes(n=47):
    while n not in [0, 45, 47, 49] and n > 45:
        print(f"Only 45, 47, 49 classes are supported, got {n}, downscaling for now")
        n -= 1

    # assert n in [45, 47, 49], "Only 45, 47, 49 classes are supported"
    lumbar_vertebra_count = 4 + (n >= 47) + (n >= 49)
    lumbar_range = range(1, lumbar_vertebra_count + 1)
    vertebra_labels = get_vertebra_labels(c_range=range(2, 8), t_range=range(1, 13), l_range=lumbar_range)
    disc_labels = get_disc_labels(vertebra_labels)

    labels = list(filter(None, sum(zip_longest(vertebra_labels, disc_labels), ())))
    return labels



from functools import lru_cache

import numpy as np
from loguru import logger

from spine_segmentation.dataloader.statistics import Statistics
from spine_segmentation.resources.paths import MEASURE_STATISTICS_PATH, SECRETS_PATH, VECTOR_TABLE_PATH


@lru_cache
def get_measure_statistics() -> Statistics:
    logger.info(f"Loading measure statistics from {MEASURE_STATISTICS_PATH}")
    measure_statistics = Statistics(MEASURE_STATISTICS_PATH)
    logger.success(f"Loaded measure statistics: {measure_statistics}")
    return measure_statistics


@lru_cache
def get_vector_table() -> np.array:
    logger.info(f"Loading vector_table from {VECTOR_TABLE_PATH}")
    vector_table = np.load(VECTOR_TABLE_PATH)["vector_table"]
    logger.success(f"Loaded measure statistics: {vector_table.shape}")
    return vector_table


def get_secret(name):
    path = SECRETS_PATH / name
    if path.exists():
        return path.read_text().strip()



from pathlib import Path
from typing import Iterable, Literal

import lightning.pytorch as pl
import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset

from spine_segmentation.datasets.augmentation import ImageAugmentation
from spine_segmentation.datasets.sample import GTFormat, Sample
from spine_segmentation.datasets.path_helper import get_potential_data_dirs
from spine_segmentation.resources.paths import NAKO_DATASET_PATH


class PatchSegmentationDataset(Dataset):
    def __init__(
        self, data_dirs: Iterable[Path], transform=None, subset_size: int = 20, mode: Literal["val", "train"] = "val"
    ):
        super().__init__()
        self.target_shape = (18, 320, 896)
        self.patch_target_shape = (18, 128, 128)
        self.data_dirs = list(data_dirs)
        self.gt_format = GTFormat(order="y-sort", separation="instance", include="wk")
        self.transform = transform
        self.mode = mode
        self.subset_size = subset_size

    def __len__(self):
        return len(self.data_dirs)

    def __getitem__(self, index):
        path = self.data_dirs[index]
        image = Sample(path, self.target_shape, gt_format=self.gt_format)
        # connected_components = numerate_objects(image.gt)
        outputs = []
        targets = []
        # while True:
        # start_at, end_at = sorted(np.random.randint(1, max(np.unique(image.gt)), size=(2,)))
        # if end_at - start_at >= 5:
        #     break

        start_at = np.random.randint(1, max(np.unique(image.gt)) - self.subset_size)
        end_at = start_at + self.subset_size
        assert end_at - start_at == self.subset_size
        assert end_at <= max(np.unique(image.gt))

        for val in np.unique(image.gt)[:50]:
            if val == 0:
                continue
            if not (start_at <= val <= end_at):
                continue
            # Average coordinate of the object
            coords = np.argwhere(image.gt == val)
            coords = np.round(coords.mean(axis=0)).astype(int)

            if self.mode == "train":
                coords += np.random.randint(-5, 5, coords.shape)

            shape = self.patch_target_shape

            # Sample self.patch_target_shape around the coordinate
            patch = image.image[
                :,  # coords[0] - shape[0] // 2 : coords[0] + shape[0] // 2,
                coords[1] - shape[1] // 2 : coords[1] + shape[1] // 2,
                coords[2] - shape[2] // 2 : coords[2] + shape[2] // 2,
            ]

            # Pad if necessary
            padding = np.subtract(shape, patch.shape)
            padding[padding < 0] = 0
            padding = padding // 2
            padded_patch = np.pad(patch, ((padding[0], padding[0]), (padding[1], padding[1]), (padding[2], padding[2])))
            padded_patch = torch.from_numpy(padded_patch)

            gt_patch = image.gt[
                :,  # coords[0] - shape[0] // 2 : coords[0] + shape[0] // 2,
                coords[1] - shape[1] // 2 : coords[1] + shape[1] // 2,
                coords[2] - shape[2] // 2 : coords[2] + shape[2] // 2,
            ]

            padded_gt_patch = np.pad(
                gt_patch, ((padding[0], padding[0]), (padding[1], padding[1]), (padding[2], padding[2]))
            )
            padded_gt_patch = (padded_gt_patch == val).astype(np.float32)
            padded_gt_patch = torch.from_numpy(padded_gt_patch)

            focus_patch = padded_patch * (padded_gt_patch + 0.5)

            if _plot_patch := False:
                from matplotlib import pyplot as plt

                plt.subplot(2, 2, 1)
                current_gt = (image.gt == val).astype(float)
                current_gt[
                    :,  # coords[0] - shape[0] // 2 : coords[0] + shape[0] // 2,
                    coords[1] - shape[1] // 2 : coords[1] + shape[1] // 2,
                    coords[2] - shape[2] // 2 : coords[2] + shape[2] // 2,
                ] += 0.5

                plt.imshow(current_gt[9])
                print(coords)

                plt.subplot(2, 2, 2)
                print(patch.shape)
                plt.imshow(patch[9])

                gt_patch = image.gt[
                    :,  # coords[0] - shape[0] // 2 : coords[0] + shape[0] // 2,
                    coords[1] - shape[1] // 2 : coords[1] + shape[1] // 2,
                    coords[2] - shape[2] // 2 : coords[2] + shape[2] // 2,
                ]

                plt.subplot(2, 2, 3)
                plt.imshow(gt_patch[9])

                plt.subplot(2, 2, 4)
                plt.imshow(focus_patch[9])
                plt.suptitle(f"Ind: {val}")
                plt.show()
                print("Shown\n")

            # outputs.append(padded_patch[None, ...])
            # outputs.append(torch.stack([padded_patch, padded_gt_patch]))
            outputs.append(focus_patch[None, ...])
            targets.append(val.astype(np.float32))
            # targets.append(val.astype(int))

        return outputs, torch.tensor(targets)


class PatchSegmentationDataModule(pl.LightningDataModule):
    def __init__(self, data_dir=NAKO_DATASET_PATH, batch_size=1, num_workers=2, subset_size: int = 20, augment=False):
        super().__init__()
        self.batch_size = batch_size
        self.data_dirs = get_potential_data_dirs(data_dir)
        train_size = int(len(self.data_dirs) * 0.9)
        train_paths = self.data_dirs[:train_size]
        val_paths = self.data_dirs[train_size:]

        self.segmentation_dataloader_kwargs = dict(
            batch_size=batch_size,
            num_workers=num_workers,
        )

        segmentation_dataset_kwargs = dict()

        augmentation = ImageAugmentation() if augment else None
        self.train_dataset = PatchSegmentationDataset(
            train_paths, augmentation, mode="train", subset_size=subset_size, **segmentation_dataset_kwargs
        )
        self.val_dataset = PatchSegmentationDataset(val_paths, subset_size=subset_size, **segmentation_dataset_kwargs)

    def train_dataloader(self):
        return DataLoader(self.train_dataset, shuffle=True, **self.segmentation_dataloader_kwargs)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, **self.segmentation_dataloader_kwargs)


if __name__ == "__main__":
    dataset = PatchSegmentationDataset(get_potential_data_dirs(NAKO_DATASET_PATH))
    dataset[0]



import random
from pathlib import Path
from typing import Literal, Union

import torch

from spine_segmentation.cli.cli import colored_tracebacks
from spine_segmentation.datasets.sample import SampleIterator, get_random_crop_slice
from spine_segmentation.inference.onnx_model import ONNXInferenceModel
from spine_segmentation.instance_separation.instance_separation import (
    get_planes_from_rois,
    mask_rare_rois,
    separate_rois_with_labels,
    separate_segmented_rois_by_planes,
)
from spine_segmentation.plotting.plot_slice import get_stats, plot_npz, single_plot
from spine_segmentation.resources.paths import TRAIN_SPLIT_CSV_PATH, VAL_SPLIT_CSV_PATH
from spine_segmentation.resources.other_paths import RAW_NAKO_DATASET_PATH
from spine_segmentation.spine_types.labels import get_labels_for_n_classes
from spine_segmentation.utils.log_dir import get_next_log_dir
from spine_segmentation.visualisation.blender.open_in_blender import open_in_blender

colored_tracebacks()

import numpy as np

from spine_segmentation.datasets.segmentation_dataset import SegmentationDataModule
from spine_segmentation.datasets.path_helper import expand_path_to_data_dirs


def get_dataloader(
    train_or_val: Literal["train", "val"], path: Union[Path, str] = None, *, bs: int = 16, crop_height: int = None
):
    args = [] if path is None else [path]
    data_module = SegmentationDataModule(
        *args,
        # add_bs_wk_as_channels=True,
        add_adjacent_slices=True,
        batch_size=bs,
        num_workers=4,
        crop_height_to_px=crop_height,
        target_shape=(None, 320, 896),
    )
    if train_or_val == "train":
        return data_module.train_dataloader()
    elif train_or_val == "val":
        return data_module.val_dataloader()


def plot_model(onnx_path):
    import netron

    netron.start(onnx_path)


def open_npz_in_blender(numpy_path):
    if isinstance(numpy_path, str) or isinstance(numpy_path, Path):
        spines = np.load(numpy_path)
    else:
        spines = numpy_path
    output = spines["instances_post_processed"]
    open_in_blender({"instances": output})


def open_npz_in_blender_separate_rois(npz, render_to_instead=None):
    # output, gt, instances = npz["segmentation"], npz["gt"], npz["instances"]
    # id_to_label = spines["id_to_label"].item()
    # discs = (instances % 2 == 0) * (instances != 0) * 2
    # separated_discs = separate_segmented_rois(discs) * 2
    # planes = get_planes_from_discs(separated_discs, id_to_label=id_to_label)
    # collection = BlenderCollection("planes", planes)

    open_in_blender({"instances": npz["instances"]}, render_to_instead=render_to_instead)
    # open_in_blender({"instances": instances, "gt": gt}, [collection], smooth_spine=False)


def open_npz_and_plot_rois(npz, plot_path):
    # spines = np.load(numpy_path)
    # output, gt, instances = spines["output"], spines["gt"], spines["instances"]
    # vertebrae = separate_segmented_rois(output == 1)
    # discs = separate_segmented_rois(output == 2)
    # instances = vertebrae * 2 + (discs * 2 + (discs != 0))
    plot_npz(npz, plot_path, slices=range(0, 16))


# def find_best_matching(vertebra_instances, vertebra_segmentation):


def post_process_instances(vertebra_instances, segmentation, plot_dir=None):
    vertebra_instances = vertebra_instances.copy()
    vertebra = segmentation == 1
    # from matplotlib import pyplot as plt

    # plt.imshow(vertebra_instances[8])
    # plt.imshow(instance_mask[8])
    # plt.imshow(vertebra[8])
    # plt.imshow(split_segmentation[8])
    # plt.show()
    vertebra_instances[~vertebra] = 0
    split_segmentation, _ = separate_rois_with_labels(segmentation)
    split_vertebra_segmentation = split_segmentation.copy()
    split_vertebra_segmentation[~vertebra] = 0
    split_vertebra_segmentation[vertebra] -= np.unique(split_vertebra_segmentation)[1] - 1
    add2 = vertebra * 2
    print(split_vertebra_segmentation.shape, np.unique(split_vertebra_segmentation))
    add2_only_s1 = (split_vertebra_segmentation == split_vertebra_segmentation.max()) * 2

    # best_i = 0
    best_vertebra_instances = split_vertebra_segmentation
    best_mse = 1e9

    plots = [(vertebra_instances, "Inst", np.zeros_like(vertebra_instances))]

    for ith, add_for_s1 in enumerate([0, add2_only_s1, add2_only_s1 * 2]):
        for i in range(26):
            curr_vertebra_instances = split_vertebra_segmentation + add2 * i + add_for_s1
            if (ith != 0 and curr_vertebra_instances.max() != 49) or curr_vertebra_instances.max() >= 50:
                continue
            # error_map = np.clip((vertebra_instances - curr_vertebra_instances) ** 2, -9, 9)
            error_map = np.abs(vertebra_instances - curr_vertebra_instances) + (
                vertebra_instances != curr_vertebra_instances
            )
            # error_map = vertebra_instances - curr_vertebra_instances
            # error_map[~vertebra] = 0
            mse = np.mean(error_map)
            plots.append((curr_vertebra_instances, f"{mse:.4f}", error_map))
            # print(
            #     f"MSE: {mse:.3f} {best_mse:.3f}",
            #     np.unique(curr_vertebra_instances),
            #     list(zip(*np.unique(vertebra_instances, return_counts=True))),
            # )
            if mse < best_mse:
                best_mse = mse
                best_vertebra_instances = curr_vertebra_instances

    if plot_dir is not None:
        import matplotlib.pyplot as plt

        # fig, axes = plt.subplots(4, (len(plots) + 3) // 4)
        fig, axes = plt.subplots(2, len(plots))
        fig.tight_layout(pad=0)

        from spine_segmentation.visualisation.color_palettes import get_alternating_palette

        id_to_labels = dict(zip(range(1, 50), get_labels_for_n_classes(49)))
        directory = get_next_log_dir()
        for i, (instances, title, error_map) in enumerate(plots):
            alternating = get_alternating_palette()
            ax = axes[0, i]
            single_plot(ax, alternating, None, instances[8, :, ::-1].T, id_to_labels, title, font_scale=0.4)
            ax = axes[1, i]
            single_plot(ax, alternating, error_map[8, :, ::-1].T != 0, None, id_to_labels, None, font_scale=0.4)

        plt.savefig(plot_dir / f"mse_comparison.png", dpi=600)
        plt.close()
    # fig.close()

    # np.unique(split_vertebra_segmentation + add2 * 100)
    # plt.imshow((vertebra_instances - (split_vertebra_segmentation + add2 * i))[10])
    # plt.imshow((split_vertebra_segmentation)[10])
    # plt.show()

    # vertebra_instances = split_vertebra_segmentation + add2 * best_i
    vertebra_instances = best_vertebra_instances

    # print(f"{np.unique(split_segmentation)=}")
    # for i in np.unique(split_segmentation)[1:]:
    #     instance_mask = split_segmentation == i
    #     instance = vertebra_instances[instance_mask]
    #     most_common = np.bincount(instance).argmax()
    #     vertebra_instances[instance_mask] = most_common

    planes = get_planes_from_rois(vertebra_instances)
    # vertebrae = separate_segmented_rois(arr3d == 1)
    disc_instances = separate_segmented_rois_by_planes((segmentation == 2).astype(int), planes, use_plane_index=True)
    disc_instances[vertebra_instances != 0] = 0

    instances = vertebra_instances + disc_instances
    return mask_rare_rois(instances, 50)


def print_stats_summary(stats):
    correct_sum, correct_no_edges_sum = 0, 0
    for stat in stats:
        if stat:
            correct_sum += stat["correct"]
            correct_no_edges_sum += stat["correct_no_edges"]
    print(f"{correct_sum=} / {len(stats)}, {correct_no_edges_sum=} / {len(stats)}")


def plot_every_val_dataset(directory=None, device=3, model_height=896, cropped_height=None, cropped_V=None):
    if directory is None:
        directory = get_next_log_dir()
    directory = Path(directory)

    if cropped_V is not None:
        assert 1 <= cropped_V <= 49

    if cropped_height is not None:
        assert 50 <= cropped_V

    onnx_seg_inference = ONNXInferenceModel.get_best_segmentation_model(device=device)
    # onnx_seg_inference.load_index_list()

    onnx_inst_inference = ONNXInferenceModel.get_best_instance_segmentation_model(device=device)

    sample_iterator = SampleIterator(
        expand_path_to_data_dirs(RAW_NAKO_DATASET_PATH), add_adjacent_slices=True, skip_first_percent=0.9505
    )

    # index_list_path = (
    #     Path.home() / "devel/src/git/spine_annotator/cache/onnx_inference/2023-09-12_Seg_Unet_mitb5/index_list.npz"
    # )
    # val_dataloader = get_dataloader("val", index_list_path, bs=1)

    val_dataloader = SegmentationDataModule(
        [TRAIN_SPLIT_CSV_PATH, VAL_SPLIT_CSV_PATH],
        add_adjacent_slices=True,
        batch_size=16,
        # crop_height_to_px=416,
        target_shape=(18, 320, 896),
    ).val_dataloader()

    stats = []

    gt = None
    # cropped_height = 50
    random.seed(0)
    # for i, (image, _, sample, path) in enumerate(sample_iterator):
    # for i, (image, *_) in enumerate(sample_iterator):
    for i, (image, gt) in enumerate(val_dataloader):
        # if i >= 100:
        # break

        if isinstance(image, torch.Tensor):
            image = image.cpu().detach().numpy()

        if gt is None:
            gt_npz = onnx_seg_inference.inference(image)
        else:
            gt_instances, id_to_label = separate_rois_with_labels(gt.numpy()[:, 0, :, :])
            gt_npz = {"instances": gt_instances, "id_to_label": id_to_label}

        gt_npz["instances"][gt_npz["instances"] == gt_npz["instances"].max()] = 49

        print(gt_npz["instances"].shape)

        orig_gt_nonzero = np.nonzero(np.any(gt_npz["instances"], axis=(0, 1)))[0]
        gt_start_at = orig_gt_nonzero[0]
        gt_end_at = orig_gt_nonzero[-1]
        print(f"{gt_start_at=} {gt_end_at=}")
        how_many_crops = 10
        if cropped_height == model_height == 896:
            how_many_crops = 1
            gt_end_at = gt_start_at = 0

        # for start_at in np.linspace(gt_start_at, gt_end_at - cropped_height, how_many_crops):
        for start_at_class in range(1, 49 - cropped_V + 2, 2):  # +2 so that 49 is inclusive
            end_at_class = start_at_class + cropped_V - 1
            try:
                print(f"\n\n============ Processing sample {i}: {start_at_class}-{end_at_class} =============\n")
                print(gt_npz["instances"].shape)
                relevant_classes = np.any(
                    (gt_npz["instances"] >= start_at_class) & (gt_npz["instances"] <= end_at_class), axis=(0, 1)
                )
                print(f"{relevant_classes.shape=}")
                nonzero_relevant_classes = np.nonzero(relevant_classes)[0]
                start_at = nonzero_relevant_classes[0]
                end_at = nonzero_relevant_classes[-1]
                cropped_height = end_at - start_at
                print(f"{start_at=} {end_at=} {cropped_height=}")

                image = np.array(image)
                print(image.shape)

                crop_slice = get_random_crop_slice(cropped_height, image.shape[3], start_at=round(start_at))

                curr_dir = directory / f"sample_{i:03}_{crop_slice.start:03}-{crop_slice.stop:03}"
                curr_dir.mkdir(parents=True, exist_ok=True)

                # ==================
                # Create ground truth segmentation
                # ==================

                cropped_instance_gt = gt_npz["instances"][:, None, :, crop_slice].copy(order="C")

                # ==================
                # Create cropped input image for segmentation model (which will be used as input for instance segmentation)
                # ==================

                cropped_seg_input = image[:, :, :, crop_slice].copy(order="C")
                padding_to_896 = ((0, 0), (0, 0), (0, 0), (0, image.shape[3] - cropped_height))
                cropped_padded_seg_input = np.pad(cropped_seg_input, padding_to_896, mode="constant")
                seg_npz = onnx_seg_inference.inference(cropped_padded_seg_input)

                inst_seg_input = seg_npz["segmentation"][:, None, :, :]
                cropped_inst_seg_input = inst_seg_input[:, :, :, :model_height]
                padding_to_416 = ((0, 0), (0, 0), (0, 0), (0, model_height - cropped_height))
                cropped_input_image_with_gt = np.concatenate(
                    [
                        np.pad(image[:, 1:2, :, crop_slice], padding_to_416, mode="constant"),
                        (cropped_inst_seg_input == 1),
                        (cropped_inst_seg_input == 2),
                    ],
                    axis=1,
                )
                cropped_input_image_with_gt = cropped_input_image_with_gt.astype(np.float32)

                cropped_input_image_without_gt = np.pad(
                    image[:, :, :, crop_slice], padding_to_416, mode="constant"
                ).astype(np.float32)

                # ==================
                # Create cropped input image for instance segmentation model
                # ==================

                # cropped_instance_seg_input = full_input_image_with_gt[:, :, :, crop_slice].copy(order="C")

                # if True:
                #     padding = ((0, 0), (0, 0), (0, 0), (0, image.shape[3] - height))
                #     cropped_image = np.pad(cropped_image, padding, mode="constant")
                #     cropped_gt = np.pad(cropped_gt, padding, mode="constant")

                # npz["instances"] = onnx_inst_inference.inference(cropped_image, gt=cropped_gt)["instances"]
                # npz["gt_instances"] = npz["instances"]
                new_npz = onnx_inst_inference.inference(
                    cropped_input_image_without_gt, gt=np.pad(cropped_instance_gt, padding_to_416, mode="constant")
                )
                new_npz["instances"] = new_npz["instances"] * 2 - (new_npz["instances"] != 0)
                new_npz["id_to_label"] = gt_npz["id_to_label"]
                new_npz["gt_id_to_label"] = gt_npz["id_to_label"]

                new_npz["instances_post_processed"] = post_process_instances(
                    new_npz["instances"], cropped_inst_seg_input[:, 0, :, :], plot_dir=curr_dir
                )

                new_npz["cropped_segmentation"] = cropped_inst_seg_input[:, 0, :, :]

                stats.append(get_stats(new_npz["instances_post_processed"], new_npz["gt_instances"]))
                np.savez_compressed(directory / "stats.npz", stats=stats)

                open_npz_in_blender(new_npz)
                plot_npz(new_npz, curr_dir / "plot.png", slices=range(7, 10), stats=stats[-1])
                print_stats_summary(stats)
                # if i > 50:
                #    break
                # exit(0)
            except KeyboardInterrupt:
                raise
            except:
                print(f"Error processing sample {i}")
                stats.append({})
                import traceback

                traceback.print_exc()
                continue

    # open_npz_and_plot_rois(new_npz, curr_dir / "plot.png")
    # open_npz_in_blender_separate_rois(npz, curr_dir / "render.png")


def main():
    # Get first argument as int
    import sys

    slice_height = int(sys.argv[1]) if len(sys.argv) > 1 else 49
    gpu = int(sys.argv[2]) if len(sys.argv) > 2 else "CPU"
    # plot_every_val_dataset(cropped_height=slice_height, device=gpu)
    plot_every_val_dataset(cropped_V=slice_height, device=gpu)


if __name__ == "__main__":
    main()



from collections import Counter
from pathlib import Path

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from spine_segmentation.resources.preloaded import get_measure_statistics

plots_path = Path(__file__).absolute().parent / "plots" / "metadata"


def main():
    plots_path.mkdir(parents=True, exist_ok=True)

    stats = get_measure_statistics()
    metadata = stats.dicom_metadata
    columns = ["PatientAge", "PatientSex", "PatientWeight", "PatientSize", "InstitutionAddress", "StudyDate"]
    for column_postfix in columns:
        column_name = f"DICOM;{column_postfix}"
        print(column_name)
        data = metadata[column_name]

        if column_postfix == "PatientAge":
            data = data.str.replace("Y", "").astype(int)

        if column_postfix == "StudyDate":
            month = ((data // 100 % 100) - 1) / 12
            year = data // 10000
            data = year + month
            print(year, month)

        try:
            average = np.average(data)
        except:
            average = None

        # try:
        #     data.plot.hist(bins=np.unique(data))
        # except TypeError:
        #     counts = pd.DataFrame.from_dict(Counter(data), orient="index")
        #     counts.plot(kind="bar")
        # title = f"Histogram of {column_name}"
        # if average is not None:
        #     title += f" ({average:.2f})"

        if data.dtype.kind in "biufc":  # Types representing continuous data
            # For continuous data, use seaborn's histplot
            plt.figure(figsize=(10, 6))
            sns.histplot(data, bins=len(np.unique(data)), kde=False)
        else:
            # For categorical data, use seaborn's countplot or barplot
            plt.figure(figsize=(10, 6))
            if isinstance(data, pd.Series):
                sns.countplot(x=data)
            else:  # handling raw lists or arrays
                data_count = pd.DataFrame.from_dict(Counter(data), orient="index").reset_index()
                data_count.columns = ["category", "count"]
                sns.barplot(x="category", y="count", data=data_count)

        # Set title and labels
        title = f"Histogram of {column_name}"
        if average is not None:
            title += f" ({average:.2f})"
        plt.title(title)
        plt.xlabel(column_name.split(";")[-1])
        plt.ylabel("Count")
        plt.tight_layout()
        plt.savefig(plots_path / f"{column_name.split(';')[-1]}Plot.png", dpi=300)
        plt.show()


if __name__ == "__main__":
    main()



import random

from spine_segmentation.cli.cli import colored_tracebacks
from spine_segmentation.instance_separation.instance_separation import separate_rois_with_labels
from spine_segmentation.resources.paths import TRAIN_SPLIT_CSV_PATH, VAL_SPLIT_CSV_PATH

colored_tracebacks()

from collections import defaultdict

import numpy as np

from spine_segmentation.datasets.segmentation_dataset import SegmentationDataModule


def volume_size():
    val_dataloader = SegmentationDataModule(
        [TRAIN_SPLIT_CSV_PATH, VAL_SPLIT_CSV_PATH],
        slice_wise=False,
        batch_size=1,
        # crop_height_to_px=416,
        target_shape=(18, 320, 896),
    ).train_dataloader()

    random.seed(0)
    nums = defaultdict(list)
    mins = defaultdict(lambda: 1e9)
    sums = defaultdict(lambda: 0)
    avg = defaultdict(lambda: 0)
    for index, (image, gt) in enumerate(val_dataloader):
        print(f"\n\n==================== Processing sample {index} =====================\n")
        print(image.shape)
        instances, labels = separate_rois_with_labels(gt.detach().cpu().numpy()[0], split_along_discs=False)
        for i, count in list(zip(*np.unique(instances, return_counts=True)))[1:]:
            l = labels.get(i, i)
            nums[l].append(count)
            mins[l] = min(mins[l], count)
            sums[l] += count
            avg[l] = sums[l] / (index + 1)
            print(
                f"{labels.get(i, 'NOO'):<8} {count:<7} {i:5=} min={mins[l]:<9} avg={avg[l]:<9.0f} sum={sums[l]:<9} {sorted(nums[l])[:5]}"
            )


def main():
    volume_size()


if __name__ == "__main__":
    main()



import numpy as np
import vispy.scene
from matplotlib import pyplot as plt
from pandas import DataFrame
from vispy.scene import visuals

from spine_segmentation.annotators.vector_annotator import VectorAnnotator
from spine_segmentation.resources.paths import PLOTS_PATH
from spine_segmentation.resources.preloaded import get_measure_statistics


def _evaluate_and_save_off_by(patient_centers):
    off_by_table = np.ones((patient_centers.shape[0], patient_centers.shape[1])) * (-1)
    annotator = VectorAnnotator()
    for i in range(patient_centers.shape[0]):
        off_by = annotator.evaluate_each_roi_on_patient(6, i)
        for j in sorted(off_by):
            off_by_table[i, j] = off_by[j]
        if i % 10 == 0:
            print(f"Saving at {i}/{patient_centers.shape[0]}")
            np.savez_compressed("classified_as_table.npz", classified_as_table=off_by_table)

    # print confusion matrix
    off_by_table = np.nan_to_num(off_by_table)
    off_by_table = off_by_table.astype(int)
    print("Confusion matrix:")
    print(DataFrame(off_by_table).value_counts().to_string())


def _create_confusion_matrix():
    from sklearn.metrics import confusion_matrix

    data = np.load("classified_as_table.npz")["classified_as_table"]

    data = data[:2901, :40]

    # Assuming you have a numpy array called 'data' with shape (2900, 47)
    classification_guesses = data  # Last column represents the classification guesses
    actual_indices = np.arange(data.shape[1])[None].repeat(
        data.shape[0], axis=0
    )  # Array of indices from 0 to (number of data points - 1)

    # Generate confusion matrix
    cm = confusion_matrix(actual_indices.flatten(), classification_guesses.flatten())

    # Plot confusion matrix
    for i, matrix in enumerate([cm, np.log(cm)]):
        plt.imshow(matrix, interpolation="nearest", cmap=plt.cm.Blues)
        plt.colorbar()
        plt.title(f"Confusion matrix {'(logarithmic)' if i == 1 else ''}")
        plt.ylabel("True label")
        plt.xlabel("Predicted label")
        plt.savefig(PLOTS_PATH / f"confusion_matrix_{i}.png")
        plt.show()

    # Print the confusion matrix
    print(cm)


def _plot_tsne_subset_evaluate(patient_centers, subset_size):
    patient_centers = patient_centers[:subset_size]
    patient_centers = np.nan_to_num(patient_centers)

    tsne_results = np.load("tsne_results.npy")
    off_by_table = np.load("off_by_table.npz")["off_by_table"]

    face_colors = []
    for i in range(subset_size):
        face_colors.extend(
            [
                {
                    -1: (0, 0, 0, 0.1),
                    0: (0, 1, 0, 0.8),
                    2: (1, 0.5, 0, 0.8),
                }.get(off_by_table[i, j], (1, 0, 0, 0.9))
                for j in range(off_by_table.shape[1])
            ]
        )
    DataFrame(tsne_results).plot.scatter(x=0, y=1, c=face_colors, s=0.3)
    plt.title(
        f"t-SNE of {int(tsne_results.shape[0] / patient_centers.shape[1])} patients, {tsne_results.shape[0]} ROIs"
    )
    labels = ["incorrect", "off by 2", "off by 0", "unknown"]
    colors = ["red", "orange", "green", "grey"]
    handles = [plt.Line2D([], [], color=color, marker="o", linestyle="None") for color in colors]
    plt.legend(handles, labels)
    plt.savefig("tsne.png", dpi=300)
    plt.show()


def _flatten_to_coordinates(separate_roi_xyz):
    """Input shape: (10k, 51, 3), output shape: (510k, 3)"""
    return separate_roi_xyz.reshape(-1, 3)


def plot_3d_subset_evaluate(patient_centers, subset_size):
    # annotator = VectorAnnotator()
    patient_centers = patient_centers[:subset_size]
    face_colors = []
    off_by_table = np.load("off_by_table.npz")["off_by_table"]
    print(off_by_table.shape)
    for i in range(off_by_table.shape[0]):
        # off_by = annotator.evaluate_each_roi_on_patient(6, i)
        # off_by = off_by_table[i]
        face_colors.extend(
            [
                {
                    -1: (1, 1, 1, 0.2),
                    0: (0, 1, 0, 0.8),
                    2: (1, 0.5, 0, 0.8),
                }.get(off_by_table[i, j], (1, 0, 0, 0.9))
                for j in range(off_by_table.shape[1])
            ]
        )
        # roi_count = patient_centers.shape[1]
        # face_colors.extend([(1, 1, 1, 0.2)] * (roi_count - len(face_colors) % roi_count))
    plot_3d(patient_centers, face_colors)


def plot_3d_with_basic_colors(patient_centers, coloring_type="alternating", vectors=None, separate=False):
    face_colors = []

    if coloring_type == "alternating":
        colors = [(1, 0, 0, 0.7) if i % 2 else (1, 1, 1, 0.7) for i in range(patient_centers.shape[1])]
        face_colors = colors * patient_centers.shape[0]

    if coloring_type == "unique_hex":
        for i in range(patient_centers.shape[0]):
            col = [1, (i // 255) / 255, (i % 255) / 255, 1]
            face_colors.extend([tuple(col)] * patient_centers.shape[1])

    if coloring_type == "single_by_hex":
        real_i = None
        for i in range(patient_centers.shape[0]):
            col = [1, (i // 255) / 255, (i % 255) / 255, 1]
            hex = 0xFF0D64
            is_same = hex / 255 == col[0] * 2**16 + col[1] * 2**8 + col[2]
            col[-1] = 1 if is_same else 0.0
            face_colors.extend([tuple(col)] * patient_centers.shape[1])
            if col[-1] == 1:
                real_i = i
        patient_centers = patient_centers[real_i : real_i + 1]
        face_colors = face_colors[real_i * patient_centers.shape[1] : (real_i + 1) * patient_centers.shape[1]]

    plot_3d(patient_centers, face_colors, vectors, separate=separate)


def plot_3d(patient_centers, face_colors, patient_vectors_list=None, separate=False):
    centers = _flatten_to_coordinates(patient_centers)
    """Centers shape: (510k, 3)"""

    if separate:
        separate_by = np.arange(patient_centers.shape[0]) * 50
        patient_centers[..., 1] += separate_by[:, np.newaxis]

    canvas = vispy.scene.SceneCanvas(keys="interactive", show=True)
    view = canvas.central_widget.add_view()
    # ArrowVisual()

    # create scatter object and fill in the data
    scatter = visuals.Markers()
    scatter.set_data(centers, edge_width=1, face_color=face_colors, size=2 + separate * 5)
    view.add(scatter)

    v1 = _flatten_to_coordinates(patient_vectors_list[0])
    v2 = _flatten_to_coordinates(patient_vectors_list[1])
    diff = np.sum((v1 - v2) ** 2, axis=1)[..., np.newaxis]

    alpha = 1 if separate else 0.5
    vector_colors = [(1, 0, 0, alpha), (1, 1, 1, alpha), (1, 1, 0, alpha)]

    if patient_vectors_list is not None:
        for arrow_colors, patient_vectors in zip(vector_colors, patient_vectors_list):
            vectors = _flatten_to_coordinates(patient_vectors)
            # Merge shape (500k, 3) with (500k, 3) to (500k, 2, 3)
            # arrows = np.concatenate([centers, centers+vectors*(1+separate*9)], axis=1).reshape(-1, 2, 3)
            arrows = np.concatenate([centers, centers + vectors * (diff * 100 + 3 * separate)], axis=1).reshape(
                -1, 2, 3
            )

            # (500k, 2, 3) to (500k, 6)
            arrows_dir = arrows.reshape(-1, 6)

            # arrow_colors = np.repeat(face_colors, 2, axis=0)
            arrow_visuals = visuals.Arrow(
                arrows,
                width=2,
                color=arrow_colors,
                arrow_size=1,
                connect="segments",
                arrow_type="stealth",
                # arrows=arrows_dir,
                arrow_color=arrow_colors,
            )
            view.add(arrow_visuals)

    view.camera = "turntable"  # or try 'arcball'

    # Camera centered at 00
    view.camera.center = (0, 0, 0)

    # add a colored 3D axis for orientation
    axis = visuals.XYZAxis(parent=view.scene)
    vispy.app.run()


def main():
    stats = get_measure_statistics()
    # _create_confusion_matrix()
    plot_3d_subset_evaluate(stats.centers, stats.all_vectors)
    # _evaluate_and_save_off_by(stats.centers)
    # plot_3d_with_basic_colors(stats.centers, "alternating", stats.all_vectors, separate=True)


if __name__ == "__main__":
    main()



import contextlib
import io
import os
import shutil
import socket
import sys

import mlflow
import yaml
from pytorch_lightning.cli import LightningCLI


def colored_tracebacks():
    from rich import traceback

    traceback.install(width=120)


def setup_mlflow():
    from spine_segmentation.resources.preloaded import get_secret

    os.environ["MLFLOW_TRACKING_USERNAME"] = get_secret("mlflow_username")
    os.environ["MLFLOW_TRACKING_PASSWORD"] = get_secret("mlflow_password")

    mlflow.set_tracking_uri(uri="<mlflow-url>")

    possible_config_paths = [arg for arg in sys.argv if arg.startswith("--config=")]
    if len(possible_config_paths) > 0:
        config_path = possible_config_paths[0].split("--config=")[1]

        with open(config_path, "r") as file:
            config = yaml.safe_load(file)
            if "experiment_name" in config["model"]["init_args"]:
                experiment_name = config["model"]["init_args"]["experiment_name"]
                mlflow.set_experiment(experiment_name)
            else:
                mlflow.set_experiment("spine_annotator")

        copy_to_path = "/tmp/config.yaml"
        shutil.copy(config_path, copy_to_path)
        mlflow.set_tag("ConfigPath", str(config_path))
        mlflow.log_artifact(copy_to_path, "config")
    else:
        mlflow.set_experiment("spine_annotator")

    mlflow.set_tag("mlflow.runName", "debug_run")
    mlflow.set_tag("Hostname", socket.gethostname())

    # mlflow.log_artifact("config.yaml")
    mlflow.pytorch.autolog(log_models=False, log_datasets=False)

    run_cli_but_only_log_config_in_mlflow()


def run_cli_but_only_log_config_in_mlflow():
    string_io = io.StringIO()
    with contextlib.redirect_stdout(string_io):
        try:
            sys.argv.append("--print_config")
            LightningCLI()
        except:
            pass
        finally:
            sys.argv.pop()
    config = string_io.getvalue()
    with open("/tmp/config_printed.yaml", "w") as file:
        file.write(config)
    mlflow.log_artifact("/tmp/config_printed.yaml", "config")


def main():
    colored_tracebacks()
    # setup_mlflow()
    LightningCLI()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Union, Dict

from loguru import logger

from spine_segmentation.inference.onnx_model import ONNXInferenceModel
from spine_segmentation.instance_separation.instance_separation import (
    get_planes_from_rois,
    mask_rare_rois,
    separate_rois_with_labels,
    separate_segmented_rois_by_planes,
)
from spine_segmentation.spine_types.labels import get_labels_for_n_classes
from spine_segmentation.utils.log_dir import get_next_log_dir
from spine_segmentation.visualisation.blender.open_in_blender import open_in_blender

import numpy as np


def post_process_instances(vertebra_instances, segmentation, plot_dir=None):
    vertebra_instances = vertebra_instances.copy()
    vertebra = segmentation == 1

    vertebra_instances[~vertebra] = 0
    split_segmentation, _ = separate_rois_with_labels(segmentation)
    split_vertebra_segmentation = split_segmentation.copy()
    split_vertebra_segmentation[~vertebra] = 0
    split_vertebra_segmentation[vertebra] -= np.unique(split_vertebra_segmentation)[1] - 1
    add2 = vertebra * 2
    print(split_vertebra_segmentation.shape, np.unique(split_vertebra_segmentation))
    add2_only_s1 = (split_vertebra_segmentation == split_vertebra_segmentation.max()) * 2

    best_vertebra_instances = split_vertebra_segmentation
    best_mse = 1e9

    plots = [(vertebra_instances, "Inst", np.zeros_like(vertebra_instances))]

    for ith, add_for_s1 in enumerate([0, add2_only_s1, add2_only_s1 * 2]):
        for i in range(26):
            curr_vertebra_instances = split_vertebra_segmentation + add2 * i + add_for_s1
            if (ith != 0 and curr_vertebra_instances.max() != 49) or curr_vertebra_instances.max() >= 50:
                continue
            # error_map = np.clip((vertebra_instances - curr_vertebra_instances) ** 2, -9, 9)
            error_map = np.abs(vertebra_instances - curr_vertebra_instances) + (
                vertebra_instances != curr_vertebra_instances
            )
            # error_map = vertebra_instances - curr_vertebra_instances
            # error_map[~vertebra] = 0
            mse = np.mean(error_map)
            plots.append((curr_vertebra_instances, f"{mse:.4f}", error_map))
            # print(
            #     f"MSE: {mse:.3f} {best_mse:.3f}",
            #     np.unique(curr_vertebra_instances),
            #     list(zip(*np.unique(vertebra_instances, return_counts=True))),
            # )
            if mse < best_mse:
                best_mse = mse
                best_vertebra_instances = curr_vertebra_instances


    # vertebra_instances = split_vertebra_segmentation + add2 * best_i
    vertebra_instances = best_vertebra_instances

    planes = get_planes_from_rois(vertebra_instances)
    # vertebrae = separate_segmented_rois(arr3d == 1)
    disc_instances = separate_segmented_rois_by_planes((segmentation == 2).astype(int), planes, use_plane_index=True)
    disc_instances[vertebra_instances != 0] = 0

    instances = vertebra_instances + disc_instances
    return mask_rare_rois(instances, 50)


def print_stats_summary(stats):
    correct_sum, correct_no_edges_sum = 0, 0
    for stat in stats:
        if stat:
            correct_sum += stat["correct"]
            correct_no_edges_sum += stat["correct_no_edges"]
    print(f"{correct_sum=} / {len(stats)}, {correct_no_edges_sum=} / {len(stats)}")


@dataclass
class SegmentationResult:
    semantic_segmentation: np.ndarray
    instance_segmentation: np.ndarray
    id_to_label: Dict[int, str]


DeviceType = Union[Literal["CPU", "GPU"], int]


def _add_channel_to_3d_mri(mri3d: np.ndarray):
    """Given a shape (20, 320, 896), convert it to (20, 3, 320, 896), where in the RGB channel adjacent slices are encoded"""
    first = np.roll(mri3d, axis=0, shift=1)
    first[0, :, :] = first[1, :, :]

    last = np.roll(mri3d, axis=0, shift=-1)
    last[-1, :, :] = last[-2, :, :]

    result = np.stack([first, mri3d, last], axis=1)
    new_shape = (mri3d.shape[0], 3, mri3d.shape[1], mri3d.shape[2])
    assert new_shape == result.shape
    return result


class SegmentationInference:

    def __init__(self, segmentation_device: DeviceType = "CPU", instance_segmentation_device: DeviceType = "CPU", *, output_same_shape_as_input: bool = False):
        self._model_image_height = 896
        self._model_image_width = 320
        self._onnx_seg_inference = ONNXInferenceModel.get_best_segmentation_model(device=segmentation_device)
        self._onnx_inst_inference = ONNXInferenceModel.get_best_instance_segmentation_model(
            device=instance_segmentation_device
        )
        self._output_same_shape_as_input = output_same_shape_as_input

    def segment(
        self,
        mri_3d_numpy_image: np.ndarray,
        *,
        batch_size: int,
        cache_dir: Union[Literal["auto"], None, Path, str] = "auto",
    ):
        initial_shape = mri_3d_numpy_image.shape
        if len(mri_3d_numpy_image.shape) < 3:
            logger.error(
                f"mri_3d_numpy_image must be 3 dimensional. Got {mri_3d_numpy_image.shape}. If you "
                f"wish to segment only a 2D slice, consider repeating the slice three times in the "
                f"first dimension, resulting in this shape: {(3, *mri_3d_numpy_image.shape)}:"
                "\n\n\tnp.tile(input_image, (3, 1, 1))\n"
            )
            exit(1)
        if len(mri_3d_numpy_image.shape) > 3:
            logger.error(f"mri_3d_numpy_image must be 3 dimensional. Got {mri_3d_numpy_image.shape}. "
                         f"Consider dropping the batch-dimension and just using the 3D mri image as entire input.")

        mri_3d_numpy_image = self._crop_and_pad_to_shape(mri_3d_numpy_image, self._model_image_width, self._model_image_height).astype(np.float32)
        mri_4d_with_channels = _add_channel_to_3d_mri(mri_3d_numpy_image)

        seg_npz = self._onnx_seg_inference.inference(mri_4d_with_channels)
        inst_seg_input = seg_npz["segmentation"][:, None, :, :]
        cropped_inst_seg_input = inst_seg_input[:, :, :, :self._model_image_height]
        cropped_input_image_with_gt = np.concatenate(
            [
                mri_4d_with_channels[:, 1:2, :, :],
                (cropped_inst_seg_input == 1),
                (cropped_inst_seg_input == 2),
            ],
            axis=1,
        ).astype(np.float32)
        cropped_input_image_without_gt = mri_4d_with_channels[:, :, :, :].astype(np.float32)

        new_npz = self._onnx_inst_inference.inference(cropped_input_image_without_gt)
        new_npz["instances"] = new_npz["instances"] * 2 - (new_npz["instances"] != 0)
        # new_npz["id_to_label"] = gt_npz["id_to_label"]
        # new_npz["gt_id_to_label"] = gt_npz["id_to_label"]
        new_npz["instances_post_processed"] = post_process_instances(
            new_npz["instances"], cropped_inst_seg_input[:, 0, :, :], plot_dir=cache_dir
        )
        new_npz["cropped_segmentation"] = cropped_inst_seg_input[:, 0, :, :]
        id_to_labels = get_labels_for_n_classes(49)

        instances = new_npz["instances_post_processed"]
        segmentation = seg_npz["segmentation"]
        if self._output_same_shape_as_input:
            segmentation = self._crop_and_pad_to_shape(segmentation, initial_shape[1], initial_shape[2])
            instances = self._crop_and_pad_to_shape(instances, initial_shape[1], initial_shape[2])
            assert segmentation.shape == initial_shape

        return SegmentationResult(segmentation, instances, id_to_labels)

    def _crop_and_pad_to_shape(self, mri_3d_with_channels: np.ndarray, target_width, target_height):
        _, width, height = mri_3d_with_channels.shape

        # Crop
        from_width = width // 2 - target_width // 2
        to_width = from_width + target_width
        mri_3d_with_channels = mri_3d_with_channels[:, from_width:to_width, 0:target_height]

        # Padding
        _, width, height = mri_3d_with_channels.shape
        # 2 different widths, because they could differ by 1 pixel
        half_width_padding = (target_width - width) // 2
        other_half_padding = target_width - width - half_width_padding
        height_padding = target_height - height
        padding_to_896 = ((0, 0), (half_width_padding, other_half_padding), (0, height_padding))
        mri_3d_with_channels = np.pad(mri_3d_with_channels, padding_to_896, mode="constant")
        assert mri_3d_with_channels.shape[1:] == (target_width, target_height), f"{mri_3d_with_channels.shape[1:]} != {(target_width, target_height)=}"
        return mri_3d_with_channels



from collections import OrderedDict
from functools import lru_cache
from typing import Dict

import seaborn as sns

from spine_segmentation.visualisation.color import Color


@lru_cache
def get_alternating_palette() -> Dict[int, Color]:
    vertebra_count = 27
    disc_count = 26

    # vertebra_palette = sns.cubehelix_palette(n_colors=26, start=0.5, rot=-0.75)
    # disc_palette = sns.color_palette("viridis", n_colors=25)
    vertebra_palette = sns.color_palette("Paired", n_colors=vertebra_count)
    disc_palette = sns.color_palette("Paired", n_colors=disc_count + 1)[:-1]
    # vertebra_palette = Color.make_gradient(Color("orange"), Color("red"), vertebra_count)
    # disc_palette = Color.make_gradient(Color("green"), Color("cyan"), disc_count)
    vertebra_palette = [Color(*c) for c in vertebra_palette]
    disc_palette = [Color(*c) for c in disc_palette]
    disc_palette = disc_palette[2:] + disc_palette[:2]

    vertebra_colors = dict(zip(range(1, 100, 2), vertebra_palette))
    disc_colors = dict(zip(range(2, 100, 2), disc_palette))
    # Swap blue and light blue in order to remove S1 (light blue) being adjacent to light blue (48)
    disc_colors[46] = disc_colors[48]
    disc_colors[48] = disc_colors[52]
    palette = {0: Color("black"), **vertebra_colors, **disc_colors}
    ordered = OrderedDict(sorted(palette.items()))
    for i in set(palette.keys()) - set(range(1, 50)):
        del ordered[i]
    return ordered


def get_wk_bs_palette() -> Dict[int, Color]:
    return OrderedDict(
        {
            0: Color("black"),
            1: Color("orange"),
            2: Color(
                0.0,
                0.5,
                0.9,
            ),
        }
    )
